# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`compute.py`
Compute utilities.
"""

import logging
import pprint
import time
from typing import Optional

from slo_generator import constants, utils
from slo_generator.migrations.migrator import report_v2tov1
from slo_generator.report import SLOReport

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-many-arguments,too-many-locals
def compute(
    slo_config: dict,
    config: dict,
    timestamp: Optional[float] = None,
    client=None,
    do_export: bool = False,
    delete: bool = False,
):
    """Run pipeline to compute SLO, Error Budget and Burn Rate, and export the
    results (if exporters are specified in the SLO config).

    Args:
        slo_config (dict): SLO configuration.
        config (dict): SLO Generator configuration.
        timestamp (float, optional): UNIX timestamp. Defaults to now.
        client (obj, optional): Existing metrics backend client.
        do_export (bool, optional): Enable / Disable export. Default: False.
        delete (bool, optional): Enable / Disable delete mode. Default: False.
    """
    start = time.time()
    if timestamp is None:
        timestamp = time.time()

    if slo_config is None:
        LOGGER.error("SLO configuration is empty")
        return []

    # Get exporters, backend and error budget policy
    spec = slo_config["spec"]
    exporters = utils.get_exporters(config, spec)
    default_exporters_spec = {"exporters": config.get("default_exporters", [])}
    default_exporters = utils.get_exporters(config, default_exporters_spec)
    exporters.extend(x for x in default_exporters if x not in exporters)
    error_budget_policy = utils.get_error_budget_policy(config, spec)
    backend = utils.get_backend(config, spec)
    reports = []
    for step in error_budget_policy["steps"]:
        report = SLOReport(
            config=slo_config,
            backend=backend,
            step=step,
            timestamp=timestamp,
            client=client,
            delete=delete,
        )
        json_report = report.to_json()

        if not report.valid:
            LOGGER.error(report)
            reports.append(json_report)
            continue

        if delete:  # delete mode is enabled
            continue

        LOGGER.info(report)
        if exporters is not None and do_export is True:
            errors = export(json_report, exporters)
            json_report["errors"].extend(errors)
        reports.append(json_report)
    end = time.time()
    run_duration = round(end - start, 1)
    LOGGER.debug(pprint.pformat(reports))
    LOGGER.info(f"Run finished successfully in {run_duration}s.")
    return reports


def export(data: dict, exporters: list, raise_on_error: bool = False) -> list:
    """Export data using selected exporters.

    Args:
        data (dict): Data to export.
        exporters (list): List of exporter configurations.

    Returns:
        list: List of export errors.
    """
    LOGGER.debug(f"Exporters: {pprint.pformat(exporters)}")
    LOGGER.debug(f"Data: {pprint.pformat(data)}")
    name = data["metadata"]["name"]
    ebp_step = data["error_budget_policy_step_name"]
    info = f"{name :<32} | {ebp_step :<8}"
    errors = []

    # Passing one exporter as a dict will work for convenience
    if isinstance(exporters, dict):
        exporters = [exporters]
    if not exporters:
        error = "No exporters were found."
        LOGGER.error(f"{info} | {error}")
        errors.append(error)

    for exporter in exporters:
        try:
            cls = exporter.get("class")
            name = exporter.get("name")
            instance = utils.get_exporter_cls(cls)
            if not instance:
                raise ImportError("Exporter not found in shared config.")
            LOGGER.debug(f"Exporter config: {pprint.pformat(exporter)}")

            # Convert data to export from v1 to v2 for backwards-compatible
            # exporters such as BigQuery.
            json_data = data
            if cls not in constants.V2_EXPORTERS:
                LOGGER.debug(f"{info} | Converting SLO report to v1.")
                json_data = report_v2tov1(data)
            LOGGER.debug(f"{info} | SLO report: {json_data}")
            response = instance().export(json_data, **exporter)
            LOGGER.info(f'{info} | SLO report sent to "{name}" exporter successfully.')
            LOGGER.debug(f"{info} | {response}")
        except Exception as exc:  # pylint: disable=broad-except
            if raise_on_error:
                raise exc
            tbk = utils.fmt_traceback(exc)
            error = f'{cls}Exporter "{name}" failed. | {tbk}'
            LOGGER.error(f"{info} | {error}")
            LOGGER.exception(exc)
            errors.append(error)
    return errors



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`cli.py`
Command-Line interface of `slo-generator`.
"""

import logging
import os
import sys
import time
from pathlib import Path

import click
from pkg_resources import get_distribution

from slo_generator import utils
from slo_generator.compute import compute as _compute
from slo_generator.constants import LATEST_MAJOR_VERSION
from slo_generator.migrations import migrator

sys.path.append(os.getcwd())  # dynamic backend loading

LOGGER = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.option(
    "--version",
    "-v",
    is_flag=True,
    help="Show slo-generator version.",
)
@click.pass_context
def main(ctx, version):
    """CLI entrypoint."""
    utils.setup_logging()
    if ctx.invoked_subcommand is None or version:
        ver = get_distribution("slo-generator").version
        print(f"slo-generator v{ver}")
        sys.exit(0)


@main.command()
@click.option(
    "--slo-config",
    "-f",
    type=click.Path(),
    required=True,
    help="SLO config path",
)
@click.option(
    "--config",
    "-c",
    type=click.Path(),
    default="config.yaml",
    show_default=True,
    help="slo-generator config path",
)
@click.option(
    "--export",
    "-e",
    is_flag=True,
    help="Export SLO report to exporters",
)
@click.option(
    "--delete",
    "-d",
    is_flag=True,
    help="Delete mode (used for backends with SLO APIs)",
)
@click.option(
    "--timestamp",
    "-t",
    type=float,
    default=time.time(),
    help="End timestamp for query.",
)
def compute(slo_config, config, export, delete, timestamp):
    """Compute SLO report."""
    start = time.time()

    # Load slo-generator config
    LOGGER.debug(f"Loading slo-generator config from {config}")
    config_dict = utils.load_config(config)

    # Load SLO config(s)
    if Path(slo_config).is_dir():
        slo_configs = utils.load_configs(slo_config, kind="ServiceLevelObjective")
    else:
        slo_configs = [utils.load_config(slo_config, kind="ServiceLevelObjective")]

    if not slo_configs:
        LOGGER.error(f"No SLO configs found in {slo_config}.")
        sys.exit(1)

    # Load SLO configs and compute SLO reports
    all_reports = {}
    for slo_config_dict in slo_configs:
        reports = _compute(
            slo_config_dict,
            config_dict,
            timestamp=timestamp,
            do_export=export,
            delete=delete,
        )
        if reports:
            name = slo_config_dict["metadata"]["name"]
            all_reports[name] = reports
    end = time.time()
    duration = round(end - start, 1)
    LOGGER.info(
        f"Run summary | SLO Configs: {len(slo_configs)} | " f"Duration: {duration}s"
    )
    LOGGER.debug(all_reports)
    return all_reports


# pylint: disable=import-error,import-outside-toplevel
@main.command()
@click.pass_context
@click.option(
    "--config",
    "-c",
    envvar="CONFIG_PATH",
    required=True,
    help="slo-generator configuration file path.",
)
@click.option(
    "--exporters",
    "-e",
    envvar="EXPORTERS",
    required=False,
    default="",
    help="List of exporters to send data to",
)
@click.option(
    "--signature-type",
    envvar="GOOGLE_FUNCTION_SIGNATURE_TYPE",
    default="http",
    type=click.Choice(["http", "cloudevent"]),
    help="Signature type",
)
@click.option(
    "--target",
    envvar="GOOGLE_FUNCTION_SIGNATURE_TYPE",
    default="run_compute",
    help="Target function name",
)
@click.option(
    "--port",
    "-p",
    default=8080,
    help="HTTP port",
)
# pylint: disable=too-many-arguments
def api(ctx, config, exporters, signature_type, target, port):
    """Run an API that can receive requests (supports both 'http' and
    'cloudevents' signature types)."""
    from functions_framework._cli import _cli

    os.environ["EXPORTERS"] = exporters
    os.environ["CONFIG_PATH"] = config
    os.environ["GOOGLE_FUNCTION_SIGNATURE_TYPE"] = signature_type
    os.environ["GOOGLE_FUNCTION_TARGET"] = target
    ctx.invoke(
        _cli,
        target=target,
        source=Path(__file__).parent / "api" / "main.py",
        signature_type=signature_type,
        port=port,
    )


@main.command()
@click.option(
    "--source",
    "-s",
    type=click.Path(exists=True, resolve_path=True, readable=True),
    required=True,
    default=Path.cwd(),
    help="Source SLO configs folder",
)
@click.option(
    "--target",
    "-t",
    type=click.Path(resolve_path=True),
    default=Path.cwd(),
    required=True,
    help="Target SLO configs folder",
)
@click.option(
    "--error-budget-policy-path",
    "-b",
    type=click.Path(exists=True, resolve_path=True, readable=True),
    required=False,
    multiple=True,
    default=["error_budget_policy.yaml"],
    help="Error budget policy path",
)
@click.option(
    "--exporters-path",
    "-e",
    type=click.Path(exists=True, resolve_path=True, readable=True),
    required=False,
    multiple=True,
    help="Exporters path",
)
@click.option(
    "--version",
    type=str,
    required=False,
    default=LATEST_MAJOR_VERSION,
    show_default=True,
    help="SLO generate major version to migrate towards",
)
@click.option(
    "--quiet",
    "-q",
    is_flag=True,
    default=False,
    help="Do not ask for user input and auto-generate config keys",
)
def migrate(**kwargs):
    """Migrate SLO configs from v1 to v2."""
    migrator.do_migrate(**kwargs)



# Copyright 2021 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`v1tov2.py`
Migrate utilities for migrating slo-generator configs from v1 to v2.
"""
# pylint: disable=line-too-long, too-many-statements, too-many-ancestors, too-many-locals, too-many-nested-blocks, unused-argument
# flake8: noqa
# pytype: skip-file
import copy
import itertools
import pprint
import random
import string
import sys
from collections import OrderedDict
from pathlib import Path

import click
from ruamel import yaml

from slo_generator import utils
from slo_generator.constants import (
    BOLD,
    CONFIG_SCHEMA,
    ENDC,
    FAIL,
    GREEN,
    METRIC_LABELS_COMPAT,
    METRIC_METADATA_LABELS_TOP_COMPAT,
    PROVIDERS_COMPAT,
    RED,
    RIGHT_ARROW,
    SLO_CONFIG_SCHEMA,
    SUCCESS,
    WARNING,
)

yaml.explicit_start = True  # type: ignore[attr-defined]
yaml.default_flow_style = None  # type: ignore[attr-defined]
yaml.preserve_quotes = True  # type: ignore[attr-defined]


# pylint: disable=too-many-arguments
def do_migrate(
    source,
    target,
    error_budget_policy_path: list,
    exporters_path: list,
    version: str,
    quiet: bool = False,
    verbose: int = 0,
):
    """Process all SLO configs in folder and generate new SLO configurations.

    Args:
        source (str): Source SLO configs folder.
        target (str): Target SLO configs folder.
        error_budget_policy_path (list): Error budget policy paths.
        exporters_path (list): Exporters paths.
        glob (str): Glob expression to add to source path.
        version (str): slo-generator major version string (e.g: v1, v2, ...)
        quiet (bool, optional): If true, do not prompt for user input.
        verbose (int, optional): Verbose level.
    """
    curver: str = "v1"
    shared_config = CONFIG_SCHEMA
    cwd = Path.cwd()
    source = Path(source).resolve()
    target = Path(target).resolve()
    source_str = source.relative_to(cwd)  # human-readable path
    target_str = target.relative_to(cwd)  # human-readable path
    ebp_paths = [Path(ebp) for ebp in error_budget_policy_path]
    exporters_paths = [Path(exp) for exp in exporters_path]

    # Create target folder if it doesn't exist
    target.mkdir(parents=True, exist_ok=True)

    # Translate error budget policy to v2 and put into shared config
    if ebp_paths:
        ebp_func = getattr(sys.modules[__name__], f"ebp_{curver}to{version}")
        ebp_func(
            ebp_paths,
            shared_config=shared_config,
            quiet=quiet,
        )

    # Translate exporters to v2 and put into shared config
    if exporters_paths:
        exporters_func = getattr(
            sys.modules[__name__], f"exporters_{curver}to{version}"
        )
        exp_keys = exporters_func(
            exporters_paths,
            shared_config=shared_config,
            quiet=quiet,
        )

    # Process SLO configs
    click.secho("=" * 50)
    click.secho(
        f"Migrating slo-generator configs to {version} ...", fg="cyan", bold=True
    )
    paths = utils.get_files(source)
    if not paths:
        click.secho(f"{FAIL} No SLO configs found in {source}", fg="red", bold=True)
        sys.exit(1)

    curver = ""
    for source_path in paths:
        if source_path in ebp_paths + exporters_paths:
            continue
        source_path_str = source_path.relative_to(cwd)
        target_path = utils.get_target_path(
            source,
            target,
            source_path,
            mkdir=True,
        )
        target_path_str = target_path.resolve().relative_to(cwd)
        slo_config_str = source_path.open().read()
        slo_config, ind, blc = yaml.util.load_yaml_guess_indent(slo_config_str)
        curver = detect_config_version(slo_config)
        if not curver:
            continue

        # Source path info
        click.secho("-" * 50)
        click.secho(f"{WARNING}{source_path_str}{ENDC} [{curver}] ")

        # If config version is same as target version, continue
        if curver == version:
            click.secho(
                f"{FAIL} {source_path_str} is already in {version} format",
                fg="red",
                bold=True,
            )
            continue

        # Create target dirs if needed
        target_path.parent.mkdir(parents=True, exist_ok=True)

        # Run vx to vy migrator method
        slo_func = getattr(sys.modules[__name__], f"slo_config_{curver}to{version}")
        slo_config_v2 = slo_func(
            slo_config,
            shared_config=shared_config,
            shared_exporters=exp_keys if exporters_paths else [],
            quiet=quiet,
        )
        if not slo_config_v2:
            continue

        # Write resulting config to target path
        extra = "(replaced)" if target_path_str == source_path_str else ""
        click.secho(f"{RIGHT_ARROW} {GREEN}{target_path_str}{ENDC} [{version}] {extra}")
        with target_path.open("w", encoding="utf8") as conf:
            yaml.round_trip_dump(
                slo_config_v2,
                conf,
                indent=ind,
                block_seq_indent=blc,
                default_flow_style=None,
            )
        click.secho(f"{SUCCESS} Success !", fg="green", bold=True)

    # Write shared config to file
    click.secho("=" * 50)
    shared_config_path = target / "config.yaml"
    shared_config_path_str = shared_config_path.relative_to(cwd)
    with shared_config_path.open("w", encoding="utf8") as conf:
        click.secho(
            f"Writing slo-generator config to {shared_config_path_str} ...",
            fg="cyan",
            bold=True,
        )
        yaml.round_trip_dump(
            shared_config,
            conf,
            Dumper=CustomDumper,
            indent=2,
            block_seq_indent=0,
            explicit_start=True,
        )
        click.secho(f"{SUCCESS} Success !", fg="green", bold=True)

    # Remove error budget policy file
    # click.secho('=' * 50)
    # click.secho(f'Removing {error_budget_policy_path} ...',
    #             fg='cyan',
    #             bold=True)
    # error_budget_policy_path.unlink()
    # click.secho(f'{SUCCESS} Success !', fg='green', bold=True)

    # Print next steps
    relative_ebp_path = ebp_paths[0].relative_to(cwd)
    click.secho("=" * 50)
    click.secho(
        f"\n{SUCCESS} Migration of `slo-generator` configs to v2 completed successfully ! Configs path: {target_str}/.\n",
        fg="green",
        bold=True,
    )
    click.secho("=" * 50)
    click.secho(
        f"{BOLD}PLEASE FOLLOW THE MANUAL STEPS BELOW TO FINISH YOUR MIGRATION:",
        fg="red",
        bold=True,
    )
    click.secho(
        f"""
    1 - Commit the updated SLO configs and your shared SLO config to version control.
    2 - [local/k8s/cloudbuild] Update your slo-generator command:
    {RED}  [-] slo-generator -f {source_str} -b {relative_ebp_path}{ENDC}
    {GREEN}  [+] slo-generator -f {target_str} -c {target_str}/config.yaml{ENDC}
    """
    )
    # 3 - [terraform] Upgrade your `terraform-google-slo` modules:
    # 3.1 - Upgrade the module `version` to 2.0.0.
    # 3.2 - Replace `error_budget_policy` field in your `slo` and `slo-pipeline` modules by `shared_config`
    # 3.3 - Replace `error_budget_policy.yaml` local variable to `config.yaml`


# pylint: disable=dangerous-default-value
def exporters_v1tov2(
    exporters_paths: list, shared_config: dict = {}, quiet: bool = False
) -> list:
    """Translate exporters to v2 and put into shared config.

    Args:
        exporters_path (list): List of exporters file paths.
        shared_config (dict): Shared config to add exporters to.
        quiet (bool): Quiet mode.

    Returns:
        list: List of exporters keys added to shared config.
    """
    exp_keys = []
    for exp_path in exporters_paths:
        with open(exp_path, encoding="utf-8") as conf:
            content = yaml.load(conf, Loader=yaml.SafeLoader)
        exporters = content

        # If exporters file has sections, concatenate all of them
        if isinstance(content, dict):
            exporters = []
            for _, value in content.items():
                exporters.extend(value)

        # If exporter not in general config, add it and add an alias for the
        # exporter. Refer to the alias in the SLO config file.
        for exporter in exporters:
            exporter = OrderedDict(exporter)
            exp_key = add_to_shared_config(
                exporter, shared_config, "exporters", quiet=quiet
            )
            exp_keys.append(exp_key)
    return exp_keys


# pylint: disable=dangerous-default-value
def ebp_v1tov2(ebp_paths: list, shared_config: dict = {}, quiet: bool = False) -> list:
    """Translate error budget policies to v2 and put into shared config

    Args:
        ebp_paths (list): List of error budget policies file paths.
        shared_config (dict): Shared config to add exporters to.
        quiet (bool): Quiet mode.

    Returns:
        list: List of error budget policies keys added to shared config.
    """
    ebp_keys = []
    for ebp_path in ebp_paths:
        with open(ebp_path, encoding="utf-8") as conf:
            error_budget_policy = yaml.load(conf, Loader=yaml.SafeLoader)
        for step in error_budget_policy:
            step["name"] = step.pop("error_budget_policy_step_name")
            step["burn_rate_threshold"] = step.pop("alerting_burn_rate_threshold")
            step["alert"] = step.pop("urgent_notification")
            step["message_alert"] = step.pop("overburned_consequence_message")
            step["message_ok"] = step.pop("achieved_consequence_message")
            step["window"] = step.pop("measurement_window_seconds")

        ebp = {"steps": error_budget_policy}
        if ebp_path.name == "error_budget_policy.yaml":
            ebp_key = "default"
        else:
            ebp_key = ebp_path.stem.replace("error_budget_policy_", "")
        ebp_key = add_to_shared_config(
            ebp,
            shared_config,
            "error_budget_policies",
            ebp_key,
            quiet=quiet,
        )
        ebp_keys.append(ebp_key)
    return ebp_keys


# pylint: disable=dangerous-default-value
def slo_config_v1tov2(
    slo_config: dict,
    shared_config: dict = {},
    shared_exporters: list = [],
    quiet: bool = False,
    verbose: int = 0,
):
    """Process old SLO config v1 and generate SLO config v2.

    Args:
        slo_config (dict): SLO Config v1.
        shared_config (dict): SLO Generator config.
        shared_exporters (list): Shared exporters keys to add to SLO configs.
        quiet (bool): If true, do not ask for user input.
        verbose (int): Verbose level.

    Returns:
        dict: SLO Config v2.
    """
    # SLO config v2 skeleton
    slo_config_v2 = OrderedDict(copy.deepcopy(SLO_CONFIG_SCHEMA))
    slo_config_v2["apiVersion"] = "sre.google.com/v2"
    slo_config_v2["kind"] = "ServiceLevelObjective"
    missing_keys = [
        key
        for key in ["service_name", "feature_name", "slo_name", "backend"]
        if key not in slo_config
    ]
    if missing_keys:
        click.secho(
            f"Invalid SLO configuration: missing key(s) {missing_keys}.", fg="red"
        )
        return None

    # Get fields from old config
    slo_metadata_name_fmt = "{service_name}-{feature_name}-{slo_name}"
    slo_metadata_name = slo_metadata_name_fmt.format(**slo_config)
    slo_description = slo_config.pop("slo_description")
    slo_target = slo_config.pop("slo_target")
    service_level_indicator = slo_config["backend"].pop("measurement", {})
    backend = slo_config["backend"]
    method = backend.pop("method")
    exporters = slo_config.get("exporters", [])
    if isinstance(exporters, dict):  # single exporter, deprecated
        exporters = [exporters]

    # Fill spec
    slo_config_v2["metadata"]["name"] = slo_metadata_name
    slo_config_v2["metadata"]["labels"] = {
        "service_name": slo_config["service_name"],
        "feature_name": slo_config["feature_name"],
        "slo_name": slo_config["slo_name"],
    }
    other_labels = {
        k: v
        for k, v in slo_config.items()
        if k not in ["service_name", "feature_name", "slo_name", "backend", "exporters"]
    }
    slo_config_v2["metadata"]["labels"].update(other_labels)
    slo_config_v2["spec"]["description"] = slo_description
    slo_config_v2["spec"]["goal"] = slo_target

    # Process backend
    backend = OrderedDict(backend)
    backend_key = add_to_shared_config(
        backend,
        shared_config,
        "backends",
        quiet=quiet,
    )
    slo_config_v2["spec"]["backend"] = backend_key
    slo_config_v2["spec"]["method"] = method

    # If exporter not in general config, add it and add an alias for the
    # exporter. Refer to the alias in the SLO config file.
    for exporter in exporters:
        exporter = OrderedDict(exporter)
        exp_key = add_to_shared_config(
            exporter, shared_config, "exporters", quiet=quiet
        )
        slo_config_v2["spec"]["exporters"].append(exp_key)

    # Add shared exporters to slo config
    for exp_key in shared_exporters:
        slo_config_v2["spec"]["exporters"].append(exp_key)

    # Fill spec
    slo_config_v2["spec"]["service_level_indicator"] = service_level_indicator

    if verbose > 0:
        pprint.pprint(dict(slo_config_v2))
    return dict(slo_config_v2)


def report_v2tov1(report: dict) -> dict:
    """Convert SLO report from v2 to v1 format, for exporters to be
    backward-compatible with v1 data format.

    Args:
        report (dict): SLO report.

    Returns:
        dict: Converted SLO report.
    """
    mapped_report: dict = {}
    for key, value in report.items():

        # If a metadata label is passed, use the metadata label mapping
        if key == "metadata":
            mapped_report["metadata"] = {}
            for subkey, subvalue in value.items():

                # v2 `metadata.labels` attributes map to `metadata` attributes
                # in v1
                if subkey == "labels":
                    labels = subvalue
                    for labelkey, labelval in labels.items():

                        # Top-level labels like 'service_name', 'feature_name',
                        # and 'slo_name'.
                        if labelkey in METRIC_METADATA_LABELS_TOP_COMPAT:
                            mapped_report[labelkey] = labelval

                        # Other labels that are mapped to 'metadata' in the v1
                        # report
                        else:
                            mapped_report["metadata"][labelkey] = labelval

                # ignore the name attribute which is just a concatenation of
                # service_name, feature_name and slo_name
                elif subkey == "name":
                    continue

                # other metadata labels are still mapped to the v1 `metadata`
                # attributes
                else:
                    mapped_report["metadata"][subkey] = subvalue

        # If a key in the default label mapping is passed, use the default
        # label mapping
        elif key in METRIC_LABELS_COMPAT:
            mapped_report[METRIC_LABELS_COMPAT[key]] = value
        else:
            mapped_report[key] = value
    return mapped_report


def get_random_suffix() -> str:
    """Get random suffix for our backends / exporters when configs clash."""
    return "".join(random.choices(string.digits, k=4))  # nosec B311


def add_to_shared_config(
    new_obj: dict, shared_config: dict, section: str, key=None, quiet: bool = False
):
    """Add an object to the shared_config.

    If the object with the same config already exists in the shared config,
    simply return its key.

    If the object does not exist in the shared config:
    * If the default key is already taken, add a random suffix to it.
    * If the default key is not taken, add the new object to the config.

    Args:
        new_obj (OrderedDict): Object to add to shared_config.
        shared_config (dict): Shared config to add object to.
        section (str): Section name in shared config to add the object under.
        key (str): Key if cannot be infered.
        quiet (bool): If True, do not ask for user input.

    Returns:
        str: Object key in the shared config.
    """
    shared_obj = shared_config[section]
    key = key or new_obj.pop("class", None)
    if not key:
        raise ValueError("Object key is undefined.")
    if "." not in key:
        key = utils.caml_to_snake(PROVIDERS_COMPAT.get(key, key))

    existing_obj = {
        k: v
        for k, v in shared_obj.items()
        if k.startswith(key.split("/")[0]) and str(v) == str(dict(new_obj))
    }
    if existing_obj:
        key = next(iter(existing_obj))
        # click.secho(f'Found existing {section} {key}')
    else:
        if key in shared_obj.keys():  # key conflicts
            if quiet:
                key += "/" + get_random_suffix()
            else:
                name = section.rstrip("s")
                cfg = pprint.pformat({key: dict(new_obj)})
                valid = False
                while not valid:
                    click.secho(
                        f"\nNew {name} found with the following config:\n{cfg}",
                        fg="cyan",
                        blink=True,
                    )
                    user_input = click.prompt(
                        f"\n{RED}{BOLD}Please give this {name} a name{ENDC}", type=str
                    )
                    former_key = key
                    key += "/" + user_input.lower()
                    if key in shared_obj.keys():
                        click.secho(
                            f'{name.capitalize()} "{key}" already exists in shared config',
                            fg="red",
                            bold=True,
                        )
                        key = former_key
                    else:
                        valid = True
                click.secho(
                    f"Backend {key} was added to shared config.", fg="green", bold=True
                )

        # click.secho(f"Adding new {section} {key}")
        shared_obj[key] = dict(new_obj)
        shared_config[section] = dict(sorted(shared_obj.items()))
    return key


def detect_config_version(config: dict) -> str:
    """Return version of an slo-generator config based on the format.

    Args:
        config (dict): slo-generator configuration.

    Returns:
        str: SLO config version.
    """
    if not isinstance(config, dict):
        click.secho(
            "Config does not correspond to any known SLO config versions.", fg="red"
        )
        return None
    api_version: str = config.get("apiVersion", "")
    kind = config.get("kind", "")
    if not kind:  # old v1 format
        return "v1"
    return api_version.split("/")[-1]


def peek(iterable):
    """Check if iterable is empty.

    Args:
        iterable (collections.Iterable): an iterable

    Returns:
        iterable (collections.Iterable): the same iterable, or None if empty.
    """
    try:
        first = next(iterable)
    except StopIteration:
        return None
    return first, itertools.chain([first], iterable)


# pylint: disable=too-few-public-methods
class CustomDumper(yaml.RoundTripDumper):
    """Dedicated YAML dumper to insert lines between top-level objects.

    Args:
        data (str): Line data.
    """

    # HACK: insert blank lines between top-level objects
    # inspired by https://stackoverflow.com/a/44284819/3786245
    # pylint: disable=missing-function-docstring
    def write_line_break(self, data: str = None):
        super().write_line_break(data)

        if len(self.indents) == 1:
            super().write_line_break()



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`stackdriver.py`
Stackdriver Monitoring exporter class.
"""
import logging

from prometheus_client import CollectorRegistry, Gauge, pushadd_to_gateway
from prometheus_client.exposition import basic_auth_handler, default_handler

from .base import MetricsExporter

LOGGER = logging.getLogger(__name__)
DEFAULT_PUSHGATEWAY_JOB = "slo-generator"


class PrometheusExporter(MetricsExporter):
    """Prometheus exporter class."""

    REQUIRED_FIELDS = ["url"]
    OPTIONAL_FIELDS = ["job", "username", "password"]

    def __init__(self):
        self.username = None
        self.password = None

    def export_metric(self, data):
        """Export data to Prometheus.

        Args:
            data (dict): Metric data.

        Returns:
            object: Prometheus API result.
        """
        self.create_timeseries(data)

    def create_timeseries(self, data):
        """Create Prometheus timeseries.

        Args:
            data (dict): Metric data.

        Returns:
            object: Metric descriptor.
        """
        name = data["name"]
        description = data["description"]
        prometheus_push_url = data["url"]
        prometheus_push_job_name = data.get("job", DEFAULT_PUSHGATEWAY_JOB)
        value = data["value"]

        # Write timeseries w/ metric labels.
        labels = data["labels"]
        registry = CollectorRegistry()
        gauge = Gauge(
            name,
            description,
            registry=registry,
            labelnames=labels.keys(),
        )
        gauge.labels(*labels.values()).set(value)

        # Handle headers
        handler = default_handler
        if "username" in data and "password" in data:
            self.username = data["username"]
            self.password = data["password"]
            handler = PrometheusExporter.auth_handler

        return pushadd_to_gateway(
            prometheus_push_url,
            job=prometheus_push_job_name,
            grouping_key=labels,
            registry=registry,
            handler=handler,
        )

    # pylint: disable=too-many-arguments
    def auth_handler(self, url, method, timeout, headers, data):
        """Handles authentication for pushing to Prometheus gateway.

        Args:
            url (str): Prometheus gateway URL.
            method (str): Prometheus query method.
            timeout (int): Prometheus timeout.
            headers (dict): Headers.
            data (dict): Data to send.

        Returns:
            func: Auth handler function.
        """
        username = self.username
        password = self.password
        return basic_auth_handler(
            url, method, timeout, headers, data, username, password
        )



# Copyright 2020 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`dynatrace.py`
Dynatrace exporter implementation.
"""
import logging
import time

from slo_generator.backends.dynatrace import DynatraceClient

from .base import MetricsExporter

LOGGER = logging.getLogger(__name__)
DEFAULT_DEVICE_ID = "slo_report"


class DynatraceExporter(MetricsExporter):
    """Backend for querying metrics from Dynatrace.

    Args:
        client (obj, optional): Existing Dynatrace client to pass.
        api_url (str): Dynatrace API URL.
        api_token (str): Dynatrace token.
    """

    METRIC_PREFIX = "custom:"
    REQUIRED_FIELDS = ["api_url", "api_token"]

    def __init__(self):
        self.client = None

    def export_metric(self, data):
        """Export SLO data to Dynatrace.

        Args:
            data (dict): Metric data.

        Returns:
            object: Dynatrace API response.
        """
        api_url, api_token = data["api_url"], data["api_token"]
        self.client = DynatraceClient(api_url, api_token)
        metric = self.get_custom_metric(data)
        code = int(metric.get("error", {}).get("code", "200"))
        if code == 404:
            LOGGER.warning("Custom metric doesn't exist. Creating it.")
            metric = self.create_custom_metric(data)
        response = self.create_timeseries(data)
        return response

    def create_timeseries(self, data):
        """Create Dynatrace timeseries.

        Args:
            data (dict): Metric data.

        Returns:
            object: Dynatrace API response.
        """
        name = data["name"]
        labels = data["labels"]
        value = data["value"]
        tags = data.get("tags", [])
        device_id = data.get("device_id", DEFAULT_DEVICE_ID)
        timestamp_ms = time.time() * 1000
        timeseries = {
            "type": DEFAULT_DEVICE_ID,
            "tags": tags,
            "properties": {},
            "series": [
                {
                    "timeseriesId": name,
                    "dimensions": labels,
                    "dataPoints": [[timestamp_ms, value]],
                }
            ],
        }
        return self.client.request(
            "post",
            "entity/infrastructure/custom",
            name=device_id,
            post_data=timeseries,
        )

    def create_custom_metric(self, data):
        """Create a metric descriptor in Dynatrace API.

        Args:
            data (dict): Metric data.

        Returns:
            obj: Dynatrace API response.
        """
        name = data["name"]
        device_ids = [data.get("device_id", DEFAULT_DEVICE_ID)]
        labelkeys = list(data["labels"].keys())
        metric = {
            "displayName": name,
            "unit": "Count",
            "dimensions": labelkeys,
            "types": device_ids,
        }
        return self.client.request(
            "put",
            "timeseries",
            name=name,
            post_data=metric,
        )

    def get_custom_metric(self, data):
        """Get a custom metric descriptor from Dynatrace API.

        Args:
            data (dict): Metric data.

        Returns:
            obj: Dynatrace API response.
        """
        name = data["name"]
        return self.client.request(
            "get",
            "timeseries",
            name=name,
        )



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`bigquery.py`
BigQuery exporter class.
"""
import io
import json
import logging
import pprint

import google.api_core
from google.cloud import bigquery  # type: ignore[attr-defined]

from slo_generator import constants

LOGGER = logging.getLogger(__name__)


class BigqueryExporter:
    """BigQuery exporter class."""

    def __init__(self):
        self.client = bigquery.Client(project="unset")

    def export(self, data, **config):
        """Export results to BigQuery.

        Args:
            data (dict): Data to export.
                service_name (str): Service name.
                feature_name (str): Feature name.
                slo_name (str): SLO name.
                timestamp_human (str): Timestamp in human-readable format.
                measurement_window_seconds (int): Measurement window (in s).

            config (dict): Exporter config.
                project_id (str): BigQuery dataset project id.
                dataset_id (str): BigQuery dataset id.
                table_id (str): BigQuery table id.

        Raises:
            BigQueryError (object): BigQuery exception object.
        """
        project_id = config["project_id"]
        dataset_id = config["dataset_id"]
        table_id = config["table_id"]
        self.client.project = project_id
        table_ref = self.client.dataset(dataset_id).table(table_id)
        schema_fields = [element["name"] for element in TABLE_SCHEMA]
        keep_fields = config.get("keep_fields", [])
        try:
            table = self.client.get_table(table_ref)
            table = self.update_schema(table_ref, keep=keep_fields)
        except google.api_core.exceptions.NotFound:
            table = self.create_table(
                project_id,
                dataset_id,
                table_id,
                schema=TABLE_SCHEMA,
            )

        # Format user metadata if needed
        json_data = {k: v for k, v in data.items() if k in schema_fields}
        metadata = json_data.get("metadata", {})
        if isinstance(metadata, dict):
            metadata_fields = [
                {
                    "key": key,
                    "value": value,
                }
                for key, value in metadata.items()
            ]
            json_data["metadata"] = metadata_fields

        # Write results to BQ table
        if constants.DRY_RUN:
            LOGGER.info(f"[DRY RUN] Writing data to BigQuery: \n{json_data}")
            return []
        LOGGER.debug(f"Writing data to BigQuery:\n{json_data}")
        results = self.client.insert_rows_json(
            table,
            json_rows=[json_data],
            retry=google.api_core.retry.Retry(deadline=30),
        )
        if results:
            raise BigQueryError(results)
        return results

    @staticmethod
    def build_schema(schema):
        """Takes a schema defined as JSON (see TABLE_SCHEMA definition below)
        and convert it to a BigQuery schema.

        Args:
            schema (list): JSON Schema.

        Returns:
            list: BigQuery schema.
        """
        final_schema = []
        for row in schema:
            subschema = []
            if "fields" in row:
                subschema = [
                    bigquery.SchemaField(
                        subrow["name"],
                        subrow["type"],
                        mode=subrow["mode"],
                    )
                    for subrow in row["fields"]
                ]
            field = bigquery.SchemaField(
                row["name"],
                row["type"],
                mode=row["mode"],
                fields=subschema,
            )
            final_schema.append(field)
        return final_schema

    def create_table(self, project_id, dataset_id, table_id, schema=None):
        """Creates a BigQuery table from a schema.

        Args:
            project_id (str): Project id.
            dataset_id (str): Dataset id.
            table_id (str): Table id to create.
            schema (dict): BigQuery table schema in JSON format.

        Returns:
            obj: BigQuery table object.
        """
        if schema is not None:
            schema = TABLE_SCHEMA
        pyschema = BigqueryExporter.build_schema(schema)
        table_name = f"{project_id}.{dataset_id}.{table_id}"
        LOGGER.info(f"Creating table {table_name}")
        LOGGER.debug(f"Table schema: {pyschema}")
        table = bigquery.Table(table_name, schema=pyschema)
        table.time_partitioning = bigquery.TimePartitioning(
            type_=bigquery.TimePartitioningType.DAY,
        )
        return self.client.create_table(table)

    # pylint: disable=dangerous-default-value
    def update_schema(self, table_ref, keep=[]):
        """Updates a BigQuery table schema if needed.

        Args:
            table_ref (str): BigQuery table reference.
            keep (list): List of fields to remote schema fields to keep.

        Returns:
            obj: BigQuery table object.
        """
        table = self.client.get_table(table=table_ref)
        iostream = io.StringIO("")
        self.client.schema_to_json(table.schema, iostream)
        existing_schema = json.loads(iostream.getvalue())
        existing_fields = [field["name"] for field in existing_schema]
        LOGGER.debug(f"Existing fields: {existing_fields}")

        # Fields in TABLE_SCHEMA to add / remove
        updated_fields = [
            field["name"] for field in TABLE_SCHEMA if field not in existing_schema
        ]
        extra_remote_fields = [
            field
            for field in existing_schema
            if field not in TABLE_SCHEMA and field["name"] in keep
        ]

        # If extra remote fields are detected in existing schema, update our
        # TABLE_SCHEMA with those
        if extra_remote_fields:
            LOGGER.info(f"Extra remote BigQuery fields: {extra_remote_fields}")
            TABLE_SCHEMA.extend(extra_remote_fields)

        # If new fields are detected in TABLE_SCHEMA, update BigQuery schema
        if updated_fields:
            LOGGER.info(f"Updated BigQuery fields: {updated_fields}")
            table.schema = BigqueryExporter.build_schema(TABLE_SCHEMA)
            if constants.DRY_RUN:
                LOGGER.info("[DRY RUN] Updating BigQuery schema.")
            else:
                LOGGER.info("Updating BigQuery schema.")
                LOGGER.debug(f"New schema: {pprint.pformat(table.schema)}")
                self.client.update_table(table, ["schema"])
        return table


class BigQueryError(Exception):
    """Exception raised whenever a BigQuery error happened.

    Args:
        errors (list): List of errors.
    """

    def __init__(self, errors):
        super().__init__(BigQueryError._format(errors))
        self.errors = errors

    @staticmethod
    def _format(errors):
        err = []
        for error in errors:
            err.extend(error["errors"])
        return json.dumps(err)


TABLE_SCHEMA = [
    {
        "name": "service_name",
        "type": "STRING",
        "mode": "REQUIRED",
    },
    {
        "name": "feature_name",
        "type": "STRING",
        "mode": "REQUIRED",
    },
    {
        "name": "slo_name",
        "type": "STRING",
        "mode": "REQUIRED",
    },
    {
        "name": "slo_target",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "slo_description",
        "type": "STRING",
        "mode": "REQUIRED",
    },
    {
        "name": "error_budget_policy_step_name",
        "type": "STRING",
        "mode": "NULLABLE",
    },
    {
        "name": "error_budget_remaining_minutes",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "consequence_message",
        "type": "STRING",
        "mode": "NULLABLE",
    },
    {
        "name": "error_budget_minutes",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "error_minutes",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "error_budget_target",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "timestamp_human",
        "type": "TIMESTAMP",
        "mode": "REQUIRED",
    },
    {
        "name": "timestamp",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "cadence",
        "type": "STRING",
        "mode": "NULLABLE",
    },
    {
        "name": "window",
        "type": "INTEGER",
        "mode": "REQUIRED",
    },
    {
        "name": "bad_events_count",
        "type": "INTEGER",
        "mode": "NULLABLE",
    },
    {
        "name": "good_events_count",
        "type": "INTEGER",
        "mode": "NULLABLE",
    },
    {
        "name": "sli_measurement",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "gap",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "error_budget_measurement",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "error_budget_burn_rate",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "alerting_burn_rate_threshold",
        "type": "FLOAT",
        "mode": "NULLABLE",
    },
    {
        "name": "alert",
        "type": "BOOLEAN",
        "mode": "NULLABLE",
    },
    {
        "name": "metadata",
        "type": "RECORD",
        "mode": "REPEATED",
        "fields": [
            {
                "name": "key",
                "type": "STRING",
                "mode": "NULLABLE",
            },
            {
                "name": "value",
                "type": "STRING",
                "mode": "NULLABLE",
            },
        ],
    },
]



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import warnings

from datadog.api import Metric, ServiceLevelObjective
from elasticsearch import Elasticsearch
from google.auth._default import _CLOUD_SDK_CREDENTIALS_WARNING
from mock import MagicMock, patch
from prometheus_http_client import Prometheus

from slo_generator.backends.dynatrace import DynatraceClient
from slo_generator.compute import compute, export
from slo_generator.exporters.base import MetricsExporter
from slo_generator.exporters.bigquery import BigQueryError

from .test_stubs import (
    CTX,
    load_fixture,
    load_sample,
    load_slo_samples,
    mock_dd_metric_query,
    mock_dd_metric_send,
    mock_dd_slo_get,
    mock_dd_slo_history,
    mock_dt,
    mock_dt_errors,
    mock_es,
    mock_prom,
    mock_sd,
    mock_ssm_client,
)

warnings.filterwarnings("ignore", message=_CLOUD_SDK_CREDENTIALS_WARNING)

CONFIG = load_sample("config.yaml", CTX)
STEPS = len(CONFIG["error_budget_policies"]["default"]["steps"])
SLO_CONFIGS_SD = load_slo_samples("cloud_monitoring", CTX)
SLO_CONFIGS_SDSM = load_slo_samples("cloud_service_monitoring", CTX)
SLO_CONFIGS_PROM = load_slo_samples("prometheus", CTX)
SLO_CONFIGS_ES = load_slo_samples("elasticsearch", CTX)
SLO_CONFIGS_DD = load_slo_samples("datadog", CTX)
SLO_CONFIGS_DT = load_slo_samples("dynatrace", CTX)
SLO_REPORT = load_fixture("slo_report_v2.json")
SLO_REPORT_V1 = load_fixture("slo_report_v1.json")
EXPORTERS = load_fixture("exporters.yaml", CTX)
BQ_ERROR = load_fixture("bq_error.json")

# Pub/Sub methods to patch
PUBSUB_MOCKS = [
    "google.cloud.pubsub_v1.gapic.publisher_client.PublisherClient.publish",
    "google.cloud.pubsub_v1.publisher.futures.Future.result",
    "google.api_core.grpc_helpers.create_channel",
]

# Service Monitoring method to patch
SSM_MOCKS = [
    "slo_generator.backends.cloud_service_monitoring.ServiceMonitoringServiceClient",
    "slo_generator.backends.cloud_service_monitoring.SSM.to_json",
]


class TestCompute(unittest.TestCase):
    maxDiff = None

    @patch(
        "google.api_core.grpc_helpers.create_channel",
        return_value=mock_sd(2 * STEPS * len(SLO_CONFIGS_SD)),
    )
    def test_compute_stackdriver(self, mock):
        for config in SLO_CONFIGS_SD:
            with self.subTest(config=config):
                compute(config, CONFIG)

    @patch(SSM_MOCKS[0], return_value=mock_ssm_client())
    @patch(SSM_MOCKS[1], return_value=MagicMock(side_effect=mock_ssm_client.to_json))
    @patch(
        "google.api_core.grpc_helpers.create_channel",
        return_value=mock_sd(2 * STEPS * len(SLO_CONFIGS_SDSM)),
    )
    def test_compute_ssm(self, *mocks):
        for config in SLO_CONFIGS_SDSM:
            with self.subTest(config=config):
                compute(config, CONFIG)

    @patch(SSM_MOCKS[0], return_value=mock_ssm_client())
    @patch(SSM_MOCKS[1], return_value=MagicMock(side_effect=mock_ssm_client.to_json))
    @patch(
        "google.api_core.grpc_helpers.create_channel",
        return_value=mock_sd(2 * STEPS * len(SLO_CONFIGS_SDSM)),
    )
    @patch(PUBSUB_MOCKS[0])
    @patch(PUBSUB_MOCKS[1])
    def test_compute_ssm_delete_export(self, *mocks):
        for config in SLO_CONFIGS_SDSM:
            with self.subTest(config=config):
                compute(config, CONFIG, delete=True, do_export=True)

    @patch.object(Prometheus, "query", mock_prom)
    def test_compute_prometheus(self):
        for config in SLO_CONFIGS_PROM:
            with self.subTest(config=config):
                compute(config, CONFIG)

    @patch.object(Elasticsearch, "search", mock_es)
    def test_compute_elasticsearch(self):
        for config in SLO_CONFIGS_ES:
            with self.subTest(config=config):
                compute(config, CONFIG)

    @patch.object(Metric, "query", mock_dd_metric_query)
    @patch.object(ServiceLevelObjective, "history", mock_dd_slo_history)
    @patch.object(ServiceLevelObjective, "get", mock_dd_slo_get)
    def test_compute_datadog(self):
        for config in SLO_CONFIGS_DD:
            with self.subTest(config=config):
                compute(config, CONFIG)

    @patch.object(DynatraceClient, "request", side_effect=mock_dt)
    def test_compute_dynatrace(self, mock):
        for config in SLO_CONFIGS_DT:
            with self.subTest(config=config):
                compute(config, CONFIG)

    @patch(PUBSUB_MOCKS[0])
    @patch(PUBSUB_MOCKS[1])
    @patch(PUBSUB_MOCKS[2])
    def test_export_pubsub(self, *mocks):
        export(SLO_REPORT, EXPORTERS[0])

    @patch("google.api_core.grpc_helpers.create_channel", return_value=mock_sd(STEPS))
    def test_export_stackdriver(self, mock):
        export(SLO_REPORT, EXPORTERS[1])

    @patch("google.cloud.bigquery.Client.get_table")
    @patch("google.cloud.bigquery.Client.create_table")
    @patch("google.cloud.bigquery.Client.update_table")
    @patch("google.cloud.bigquery.Client.insert_rows_json", return_value=[])
    def test_export_bigquery(self, *mocks):
        export(SLO_REPORT, EXPORTERS[2])

    @patch("google.cloud.bigquery.Client.get_table")
    @patch("google.cloud.bigquery.Client.create_table")
    @patch("google.cloud.bigquery.Client.update_table")
    @patch("google.cloud.bigquery.Client.insert_rows_json", return_value=BQ_ERROR)
    def test_export_bigquery_error(self, *mocks):
        with self.assertRaises(BigQueryError):
            export(SLO_REPORT, EXPORTERS[2], raise_on_error=True)

    @patch("prometheus_client.push_to_gateway")
    def test_export_prometheus(self, mock):
        export(SLO_REPORT, EXPORTERS[3])

    def test_export_prometheus_self(self):
        export(SLO_REPORT, EXPORTERS[7])

    @patch.object(Metric, "send", mock_dd_metric_send)
    def test_export_datadog(self):
        export(SLO_REPORT, EXPORTERS[4])

    @patch.object(DynatraceClient, "request", side_effect=mock_dt)
    def test_export_dynatrace(self, mock):
        export(SLO_REPORT, EXPORTERS[5])

    @patch.object(DynatraceClient, "request", side_effect=mock_dt_errors)
    def test_export_dynatrace_error(self, mock):
        responses = export(SLO_REPORT, EXPORTERS[5])
        codes = [r[0]["response"]["error"]["code"] for r in responses]
        self.assertTrue(all(code == 429 for code in codes))

    def test_metrics_exporter_build_data_labels(self):
        exporter = MetricsExporter()
        data = SLO_REPORT_V1
        labels = ["service_name", "slo_name", "metadata"]
        result = exporter.build_data_labels(data, labels)
        expected = {
            "service_name": SLO_REPORT_V1["service_name"],
            "slo_name": SLO_REPORT_V1["slo_name"],
            "env": SLO_REPORT_V1["metadata"]["env"],
            "team": SLO_REPORT_V1["metadata"]["team"],
        }
        self.assertEqual(result, expected)

    @patch("google.api_core.grpc_helpers.create_channel", return_value=mock_sd(STEPS))
    @patch("google.cloud.bigquery.Client.get_table")
    @patch("google.cloud.bigquery.Client.create_table")
    @patch("google.cloud.bigquery.Client.update_table")
    @patch("google.cloud.bigquery.Client.insert_rows_json", return_value=BQ_ERROR)
    def test_export_multiple_error(self, *mocks):
        exporters = [EXPORTERS[1], EXPORTERS[2]]
        errors = export(SLO_REPORT, exporters)
        self.assertEqual(len(errors), 1)
        self.assertIn("BigQueryError", errors[0])

    @patch("google.api_core.grpc_helpers.create_channel", return_value=mock_sd(STEPS))
    @patch("google.cloud.bigquery.Client.get_table")
    @patch("google.cloud.bigquery.Client.create_table")
    @patch("google.cloud.bigquery.Client.update_table")
    @patch("google.cloud.bigquery.Client.insert_rows_json", return_value=BQ_ERROR)
    def test_export_multiple_error_raise(self, *mocks):
        exporters = [EXPORTERS[1], EXPORTERS[2]]
        with self.assertRaises(BigQueryError):
            export(SLO_REPORT, exporters, raise_on_error=True)

    def test_export_wrong_class(self):
        exporters = [{"class": "Unknown"}]
        with self.assertRaises(ImportError):
            export(SLO_REPORT, exporters, raise_on_error=True)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`cloud_monitoring.py`
Cloud Monitoring exporter class.
"""
import logging

from google.cloud import monitoring_v3

from .base import MetricsExporter

LOGGER = logging.getLogger(__name__)


class CloudMonitoringExporter(MetricsExporter):
    """Cloud Monitoring exporter class."""

    METRIC_PREFIX = "custom.googleapis.com/"
    REQUIRED_FIELDS = ["project_id"]

    def __init__(self):
        self.client = monitoring_v3.MetricServiceClient()

    def export_metric(self, data: dict):
        """Export metric to Cloud Monitoring.

        Args:
            data (dict): Data to send to Cloud Monitoring.

        Returns:
            object: Cloud Monitoring API result.
        """
        self.create_timeseries(data)

    def create_timeseries(self, data: dict):
        """Create Cloud Monitoring timeseries.

        Args:
            data (dict): Metric data.

        Returns:
            object: Metric descriptor.
        """
        series = monitoring_v3.TimeSeries()
        series.metric.type = data["name"]
        series.resource.type = "global"
        labels = data["labels"]
        for key, value in labels.items():
            series.metric.labels[key] = value  # pylint: disable=E1101

        # Define end point timestamp.
        timestamp = data["timestamp"]
        seconds = int(timestamp)
        nanos = int((timestamp - seconds) * 10**9)
        interval = monitoring_v3.TimeInterval(
            {
                "end_time": {
                    "seconds": seconds,
                    "nanos": nanos,
                }
            }
        )

        # Create a new data point and set the metric value.
        point = monitoring_v3.Point(
            {
                "interval": interval,
                "value": {
                    "double_value": data["value"],
                },
            }
        )
        series.points = [point]

        # Record the timeseries to Cloud Monitoring.
        project = self.client.common_project_path(data["project_id"])
        self.client.create_time_series(name=project, time_series=[series])
        # pylint: disable=E1101
        labels = series.metric.labels
        LOGGER.debug(
            f"timestamp: {timestamp}"
            f"value: {point.value.double_value}"
            f"{labels['service_name']}-{labels['feature_name']}-"
            f"{labels['slo_name']}-{labels['error_budget_policy_step_name']}"
        )
        # pylint: enable=E1101



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`prometheus.py`
Prometheus backend implementation.
"""

import json
import logging
import os
import pprint
from typing import Dict, List, Optional, Tuple

from prometheus_http_client import Prometheus

from slo_generator.constants import NO_DATA

LOGGER = logging.getLogger(__name__)


class PrometheusBackend:
    """Backend for querying metrics from Prometheus."""

    def __init__(self, client=None, url=None, headers=None):
        self.client = client
        if not self.client:
            if url:
                os.environ["PROMETHEUS_URL"] = url
            if headers:
                os.environ["PROMETHEUS_HEAD"] = json.dumps(headers)
            self.client = Prometheus()

    def query_sli(self, timestamp, window, slo_config):
        """Query SLI value from a given PromQL expression.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            float: SLI value.
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        expr = measurement["expression"]
        response = self.query(expr, window, timestamp, operators=[])
        sli_value = PrometheusBackend.count(response)
        LOGGER.debug(f"SLI value: {sli_value}")
        return sli_value

    def good_bad_ratio(self, timestamp, window, slo_config):
        """Compute good bad ratio from two metric filters.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Note:
            At least one of `filter_bad` or `filter_valid` is required.

        Returns:
            tuple: A tuple of (good_count, bad_count).
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        good = measurement["filter_good"]
        bad = measurement.get("filter_bad")
        valid = measurement.get("filter_valid")
        operators = measurement.get("operators", ["increase", "sum"])

        # Replace window by its value in the error budget policy step
        res = self.query(good, window, timestamp, operators)
        good_count = PrometheusBackend.count(res)

        if bad:
            res = self.query(bad, window, timestamp, operators)
            bad_count = PrometheusBackend.count(res)
        elif valid:
            res = self.query(valid, window, timestamp, operators)
            valid_count = PrometheusBackend.count(res)
            bad_count = valid_count - good_count
        else:
            raise Exception("`filter_bad` or `filter_valid` is required.")

        LOGGER.debug(f"Good events: {good_count} | " f"Bad events: {bad_count}")

        return (good_count, bad_count)

    # pylint: disable=unused-argument
    def distribution_cut(
        self, timestamp: int, window: int, slo_config: dict
    ) -> Tuple[float, float]:
        """Query events for distributions (histograms).

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            float: SLI value.
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        expr = measurement["expression"]
        threshold_bucket = measurement["threshold_bucket"]
        labels = {"le": threshold_bucket}
        res_good = self.query(
            expr,
            window,
            operators=["increase", "sum"],
            labels=labels,
        )
        good_count = PrometheusBackend.count(res_good)

        # We use the _count metric to figure out the 'valid count'.
        # Trying to get the valid count from the _bucket metric query is hard
        # due to Prometheus 'le' syntax that doesn't have the alternative 'ge'
        # See https://github.com/prometheus/prometheus/issues/2018.
        expr_count = expr.replace("_bucket", "_count")
        res_valid = self.query(
            expr_count,
            window,
            operators=["increase", "sum"],
        )
        valid_count = PrometheusBackend.count(res_valid)
        bad_count = valid_count - good_count
        LOGGER.debug(f"Good events: {good_count} | " f"Bad events: {bad_count}")
        return (good_count, bad_count)

    # pylint: disable=unused-argument,redefined-builtin,dangerous-default-value
    # pylint: disable=too-many-arguments
    def query(
        self,
        filter: str,
        window: int,
        timestamp: Optional[int] = None,
        operators: list = [],
        labels: dict = {},
    ) -> dict:
        """Query Prometheus server.

        Args:
            filter (str): Query filter.
            window (int): Window (in seconds).
            timestamp (int): UNIX timestamp.
            operators (list): List of PromQL operators to apply on query.
            labels (dict): Labels dict to add to existing query.

        Returns:
            dict: Response.
        """
        filter = PrometheusBackend._fmt_query(filter, window, operators, labels)
        LOGGER.debug(f"Query: {filter}")
        response = self.client.query(metric=filter)
        response = json.loads(response)
        LOGGER.debug(pprint.pformat(response))
        return response

    @staticmethod
    def count(response: dict) -> float:
        """Count events in Prometheus response.
        Args:
            response (dict): Prometheus query response.
        Returns:
            int: Event count.
        """
        # Note: this function could be replaced by using the `count_over_time`
        # function that Prometheus provides.
        try:
            return float(response["data"]["result"][0]["value"][1])
        except (IndexError, KeyError) as exception:
            LOGGER.warning("Couldn't find any values in timeseries response.")
            LOGGER.debug(exception, exc_info=True)
            return NO_DATA  # no events in timeseries

    @staticmethod
    # pylint: disable=dangerous-default-value
    def _fmt_query(
        query: str, window: int, operators: List[str] = [], labels: Dict[str, str] = {}
    ) -> str:
        """Format Prometheus query:

        * If the PromQL expression has a `window` placeholder, replace it by the
        current window. Otherwise, append it to the expression.

        * If operators are defined, apply them to the metric in sequential
        order.

        * If labels are defined, append them to existing labels.

        Args:
            query (str): Original query in YAML config.
            window (int): Query window (in seconds).
            operators (list): Operators to wrap query with.
            labels (dict): Labels dict to add to existing query.

        Returns:
            str: Formatted query.
        """
        query = query.strip()
        if "[window" in query:
            query = query.replace("[window", f"[{window}s")
        else:
            query += f"[{window}s]"
        for operator in operators:
            query = f"{operator}({query})"
        for key, value in labels.items():
            query = query.replace("}", f', {key}="{value}"}}')
        return query



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`custom_backend.py`
Dummy sample of a custom backend.
"""
import logging

LOGGER = logging.getLogger(__name__)


class CustomBackend:
    """Custom backend that always return an SLI of 0.999."""

    def __init__(self, client=None, **kwargs):
        pass

    # pylint: disable=unused-argument
    def good_bad_ratio(self, timestamp, window, slo_config):
        """Good bad ratio method.

        Args:
            timestamp (str): UNIX timestamp.
            window (str): Window (in seconds).
            slo_config (dict): SLO Config.

        Returns:
            tuple: Good events count / bad events count.
        """
        return 100000, 100

    # pylint: disable=unused-argument,missing-function-docstring
    def query_sli(self, timestamp, window, slo_config):
        return 0.999



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`report.py`
Report utilities.
"""

import logging
from dataclasses import asdict, dataclass, field, fields
from typing import List

from slo_generator import utils
from slo_generator.constants import COLORED_OUTPUT, MIN_VALID_EVENTS, NO_DATA, Colors

LOGGER = logging.getLogger(__name__)


@dataclass(init=False)
class SLOReport:
    """SLO report dataclass. Compute an SLO report out of an SLO config and an
    Error Budget Policy step.

    Args:
        config (dict): SLO configuration.
        backend (dict): Backend configuration.
        step (dict): Error budget policy step configuration.
        timestamp (int): Timestamp.
        client (obj): Existing backend client.
        delete (bool): Backend delete action.
    """

    # pylint: disable=too-many-instance-attributes

    # SLO
    name: str
    description: str
    goal: str
    backend: str

    # SLI
    gap: float

    # Error budget
    error_budget_policy_step_name: str
    error_budget_target: float
    error_budget_measurement: float
    error_budget_burn_rate: float
    error_budget_burn_rate_threshold: float
    error_budget_minutes: float
    error_budget_remaining_minutes: float
    error_minutes: float

    # Data validation
    valid: bool

    # Global (from error budget policy)
    timestamp: int
    timestamp_human: str
    window: int
    alert: bool

    consequence_message: str

    # SLO
    exporters: list = field(default_factory=list)
    error_budget_policy: str = "default"

    # SLI
    sli_measurement: float = 0
    events_count: int = 0
    bad_events_count: int = 0
    good_events_count: int = 0

    # Metadata
    metadata: dict = field(default_factory=dict)

    # Data validation
    errors: List[str] = field(default_factory=list)

    # pylint: disable=too-many-arguments
    def __init__(self, config, backend, step, timestamp, client=None, delete=False):

        # Init dataclass fields from SLO config and Error Budget Policy
        spec = config["spec"]
        self.exporters = []
        self.__set_fields(
            **spec,
            **step,
            lambdas={
                "goal": float,
                "step": int,
                "error_budget_burn_rate_threshold": float,
            },
        )
        # Set other fields
        self.metadata = config["metadata"]
        self.timestamp = int(timestamp)
        self.name = self.metadata["name"]
        self.error_budget_policy_step_name = step["name"]
        self.error_budget_burn_rate_threshold = float(step["burn_rate_threshold"])
        self.timestamp_human = utils.get_human_time(timestamp)
        self.valid = True
        self.errors = []

        # Get backend results
        data = self.run_backend(config, backend, client=client, delete=delete)
        if not self._validate(data):
            self.valid = False
            return

        # Build SLO report
        self.build(step, data)

        # Post validation
        if not self._post_validate():
            self.valid = False

    def build(self, step, data):
        """Compute all data necessary to build the SLO report.

        Args:
            step (dict): Error Budget Policy step configuration.
            data (obj): Backend data.

        See https://landing.google.com/sre/workbook/chapters/implementing-slos/
        for details on the calculations.
        """
        LOGGER.debug(f"{self.info} | SLO report starting ...")

        # SLI, Good count, Bad count, Gap from backend results
        sli, good_count, bad_count = self.get_sli(data)
        gap = sli - self.goal

        # Error Budget calculations
        eb_target = 1 - self.goal
        eb_value = 1 - sli
        eb_remaining_minutes = self.window * gap / 60
        eb_target_minutes = self.window * eb_target / 60
        eb_minutes = self.window * eb_value / 60
        if eb_target == 0:
            eb_burn_rate = 0
        else:
            eb_burn_rate = round(eb_value / eb_target, 1)

        # Alert boolean on burn rate excessive speed.
        alert = eb_burn_rate > self.error_budget_burn_rate_threshold

        # Manage alerting message.
        if alert:
            consequence_message = step["message_alert"]
        elif eb_burn_rate <= 1:
            consequence_message = step["message_ok"]
        else:
            consequence_message = (
                "Missed for this measurement window, but not enough to alert"
            )

        # Set fields in dataclass.
        self.__set_fields(
            sli_measurement=sli,
            good_events_count=int(good_count),
            bad_events_count=int(bad_count),
            events_count=int(good_count + bad_count),
            gap=gap,
            error_budget_target=eb_target,
            error_budget_measurement=eb_value,
            error_budget_burn_rate=eb_burn_rate,
            error_budget_remaining_minutes=eb_remaining_minutes,
            error_budget_minutes=eb_target_minutes,
            error_minutes=eb_minutes,
            alert=alert,
            consequence_message=consequence_message,
        )

    def run_backend(self, config, backend, client=None, delete=False):
        """Get appropriate backend method from SLO configuration and run it on
        current SLO config and Error Budget Policy step.

        Args:
            config (dict): SLO configuration.
            backend (dict): Backend configuration.
            client (obj, optional): Backend client initiated beforehand.
            delete (bool, optional): Set to True if we're running a delete
                action.

        Returns:
            obj: Backend data.
        """
        # Grab backend class and method dynamically.
        cls_name = backend.get("class")
        method = config["spec"]["method"]
        excluded_keys = ["class", "service_level_indicator", "name"]
        backend_cfg = {k: v for k, v in backend.items() if k not in excluded_keys}
        cls = utils.get_backend_cls(cls_name)
        if not cls:
            LOGGER.warning(f"{self.info} | Backend {cls_name} not loaded.")
            self.valid = False
            return None
        instance = cls(client=client, **backend_cfg)
        method = getattr(instance, method)
        LOGGER.debug(
            f"{self.info} | "
            f"Using backend {cls_name}.{method.__name__} (from "
            f"SLO config file)."
        )

        # Delete mode activation.
        if delete and hasattr(instance, "delete"):
            method = instance.delete
            LOGGER.info(f"{self.info} | Delete mode enabled.")

        # Run backend method and return results.
        try:
            data = method(self.timestamp, self.window, config)
            LOGGER.debug(f"{self.info} | Backend response: {data}")
        except Exception as exc:  # pylint:disable=broad-except
            self.errors.append(utils.fmt_traceback(exc))
            return None
        return data

    def get_sli(self, data):
        """Compute SLI value and good / bad counts from the backend result.

        Some backends (e.g: Prometheus) are computing and returning the SLI
        value directly, others are sending a tuple (good_count, bad_count) and
        SLI value is computed from there.

        Args:
            data (obj): Backend data.

        Returns:
            tuple: A tuple of 3 values to unpack (float, int, int).
                float: SLI value.
                int: Good events count.
                int: Bad events count.

        Raises:
            Exception: When the backend does not return a proper result.
        """
        if isinstance(data, tuple):  # good, bad count
            good_count, bad_count = data
            if good_count == NO_DATA:
                good_count = 0
            if bad_count == NO_DATA:
                bad_count = 0
            LOGGER.debug(f"{self.info} | Good: {good_count} | Bad: {bad_count}")
            sli_measurement = round(good_count / (good_count + bad_count), 6)
        else:  # sli value
            sli_measurement = round(data, 6)
            good_count, bad_count = NO_DATA, NO_DATA
        return sli_measurement, good_count, bad_count

    def to_json(self) -> dict:
        """Serialize dataclass to JSON."""
        if not self.valid:
            ebp_name = self.error_budget_policy_step_name
            return {
                "metadata": self.metadata,
                "errors": self.errors,
                "error_budget_policy_step_name": ebp_name,
                "valid": self.valid,
            }
        return asdict(self)

    # pylint: disable=too-many-return-statements
    def _validate(self, data) -> bool:
        """Validate backend results. Invalid data will result in SLO report not
        being built.

        Args:
            data (obj): Backend result (expecting tuple, float, or int).

        Returns:
            bool: True if data is valid, False
        """
        # Backend not found
        if data is None:
            return False

        # Backend result is the wrong type
        if not isinstance(data, (tuple, float, int)):
            error = (
                f"Backend method returned an object of type "
                f"{type(data).__name__}. It should instead return a tuple "
                "(good_count, bad_count) or a numeric SLI value (float / int)."
            )
            self.errors.append(error)
            return False

        # Good / Bad tuple
        if isinstance(data, tuple):

            # Tuple length should be 2
            if len(data) != 2:
                error = (
                    f"Backend method returned a tuple with {len(data)} items."
                    f"Expected 2 items."
                )
                self.errors.append(error)
                return False
            good, bad = data

            # Tuple should contain only elements of type int or float
            if not all(isinstance(n, (float, int)) for n in data):
                error = (
                    "Backend method returned a tuple with some elements having"
                    " a type different than float or int"
                )
                self.errors.append(error)
                return False

            # Tuple should not contain any element with value None.
            if good is None or bad is None:
                error = (
                    f"Backend method returned a valid tuple {data} but one of "
                    "the values is None."
                )
                self.errors.append(error)
                return False

            # Tuple should not have NO_DATA everywhere
            if (good, bad) == (NO_DATA, NO_DATA):
                error = (
                    f"Backend method returned a valid tuple {data} but the "
                    "good and bad count is NO_DATA (-1)."
                )
                self.errors.append(error)
                return False

            # Tuple should not have elements where the sum is inferior to our
            # minimum valid events threshold
            if (good + bad) < MIN_VALID_EVENTS:
                error = (
                    f"Not enough valid events ({good + bad}) found. Minimum "
                    f"valid events: {MIN_VALID_EVENTS}."
                )
                self.errors.append(error)
                return False

        # Check backend float / int value
        if isinstance(data, (float, int)) and data == NO_DATA:
            error = "Backend returned NO_DATA (-1)."
            self.errors.append(error)
            return False

        # Check backend None
        if data is None:
            error = "Backend returned None."
            self.errors.append(error)
            return False

        return True

    def _post_validate(self) -> bool:
        """Validate report after build."""

        # SLI measurement should be 0 <= x <= 1
        if not 0 <= self.sli_measurement <= 1:
            error = f"SLI is not between 0 and 1 (value = {self.sli_measurement})"
            self.errors.append(error)
            return False

        return True

    # pylint: disable=dangerous-default-value
    def __set_fields(self, lambdas={}, **kwargs):
        """Set all fields in dataclasses from configs passed and apply function
        on values whose key match one in the dictionaries.

        Args:
            lambdas (dict): Dict {key: function} to apply a function on certain
            kwargs (dict): Dict of key / values to set in dataclass.
        """
        names = set(f.name for f in fields(self))
        for name in names:
            if name not in kwargs:
                continue
            value = kwargs[name]
            if name in lambdas:
                value = lambdas[name](value)
            setattr(self, name, value)

    @property
    def info(self) -> str:
        """Step information."""
        return f"{self.name :<32} | {self.error_budget_policy_step_name :<8}"

    def __str__(self) -> str:
        report = self.to_json()
        if not self.valid:
            errors_str = " | ".join(self.errors)
            return f"{self.info} | {errors_str}"
        goal_per = self.goal * 100
        sli_per = round(self.sli_measurement * 100, 6)
        gap = round(self.gap * 100, 2)
        gap_str = str(gap)
        if gap >= 0:
            gap_str = f"+{gap}"

        sli_str = f"SLI: {sli_per:<7} % | SLO: {goal_per} % | " f"Gap: {gap_str:<6}%"
        result_str = (
            "BR: {error_budget_burn_rate:<2} / "
            "{error_budget_burn_rate_threshold} | "
            "Alert: {alert:<1} | Good: {good_events_count:<8} | "
            "Bad: {bad_events_count:<8}"
        ).format_map(report)
        full_str = f"{self.info} | {sli_str} | {result_str}"
        if COLORED_OUTPUT == 1:
            if self.alert:
                full_str = Colors.FAIL + full_str + Colors.ENDC
            else:
                full_str = Colors.OKGREEN + full_str + Colors.ENDC
        return full_str



# Copyright 2021 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`main.py`
Functions Framework API (Flask).
See https://github.com/GoogleCloudPlatform/functions-framework-python for
details on the Functions Framework.
"""

import base64
import json
import logging
import os
import pprint

import requests
from flask import jsonify, make_response

from slo_generator.compute import compute, export
from slo_generator.utils import get_exporters, load_config, setup_logging

CONFIG_PATH = os.environ["CONFIG_PATH"]
LOGGER = logging.getLogger(__name__)
TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
API_SIGNATURE_TYPE = os.environ["GOOGLE_FUNCTION_SIGNATURE_TYPE"]
setup_logging()


def run_compute(request):
    """Run slo-generator compute function. Can be configured to export data as
    well, using the `exporters` key of the SLO config.

    Args:
        request (cloudevent.CloudEvent, flask.Request): Request object.

    Returns:
        list: List of SLO reports.
    """
    # Get slo-generator config
    LOGGER.info(f"Loading slo-generator config from {CONFIG_PATH}")
    config = load_config(CONFIG_PATH)

    # Process request
    data = process_req(request)
    batch_mode = request.args.get("batch", False)
    if batch_mode:
        if not API_SIGNATURE_TYPE == "http":
            raise ValueError(
                'Batch mode works only when --signature-type is set to "http".'
            )
        process_batch_req(request, data, config)
        return jsonify([])

    # Load SLO config
    slo_config = load_config(data)

    # Compute SLO report
    LOGGER.debug(f"Config: {pprint.pformat(config)}")
    LOGGER.debug(f"SLO Config: {pprint.pformat(slo_config)}")
    reports = compute(
        slo_config,
        config,
        client=None,
        do_export=True,
    )
    if API_SIGNATURE_TYPE == "http":
        reports = jsonify(reports)
    return reports


def run_export(request):
    """Run slo-generator export function. Get the SLO report data from a request
    object.

    Args:
        request (cloudevent.CloudEvent, flask.Request): Request object.

    Returns:
        list: List of SLO reports.
    """
    # Get export data
    data = process_req(request)
    slo_report = load_config(data)
    if not slo_report:
        return make_response(
            {
                "error": "SLO report is empty.",
            }
        )

    # Get SLO config
    LOGGER.info(f"Loading slo-generator config from {CONFIG_PATH}")
    config = load_config(CONFIG_PATH)

    # Construct exporters block
    spec = {}
    # pytype: disable=attribute-error
    # pylint: disable=fixme
    # FIXME `load_config()` returns `Optional[dict]` so `config` can be `None`
    default_exporters = config.get("default_exporters", [])
    # pytype: enable=attribute-error
    cli_exporters = os.environ.get("EXPORTERS", None)
    if cli_exporters:
        cli_exporters = cli_exporters.split(",")
    if not default_exporters and not cli_exporters:
        error = (
            "No default exporters set for `default_exporters` in shared config "
            f"at {CONFIG_PATH}; and --exporters was not passed to the CLI."
        )
        return make_response(
            {
                "error": error,
            },
            500,
        )
    if cli_exporters:
        spec = {"exporters": cli_exporters}
    else:
        spec = {"exporters": default_exporters}
    exporters = get_exporters(config, spec)

    # Export data
    errors = export(slo_report, exporters)
    if API_SIGNATURE_TYPE == "http":
        return jsonify(
            {
                "errors": errors,
            }
        )

    return errors


def process_req(request):
    """Process incoming request.

    Args:
        request (cloudevent.CloudEvent, flask.Request): Request object.

    Returns:
        str: Message content.
    """
    if API_SIGNATURE_TYPE == "cloudevent":
        LOGGER.info(f'Loading config from Cloud Event "{request["id"]}"')
        if "message" in request.data:  # PubSub enveloppe
            LOGGER.info("Unwrapping Pubsub enveloppe")
            content = base64.b64decode(request.data["message"]["data"])
            data = str(content.decode("utf-8")).strip()
        else:
            data = str(request.data)
    elif API_SIGNATURE_TYPE == "http":
        data = str(request.get_data().decode("utf-8"))
        LOGGER.info("Loading config from HTTP request")
        json_data = convert_json(data)
        if json_data and "message" in json_data:  # PubSub enveloppe
            LOGGER.info("Unwrapping Pubsub enveloppe")
            content = base64.b64decode(json_data["message"]["data"])
            data = str(content.decode("utf-8")).strip()
    LOGGER.debug(data)
    return data


def convert_json(data):
    """Convert string to JSON if possible or return None otherwise.

    Args:
        data (str): Data.

    Returns:
        dict: Loaded dict.
    """
    try:
        return json.loads(data)
    except ValueError:
        return None


def process_batch_req(request, data, config):
    """Process batch request. Split list of ;-delimited URLs and make one
    request per URL.

    Args:
        request (cloudevent.CloudEvent, flask.Request): Request object.
        data (str): Incoming data.
        config (dict): SLO generator config.

    Returns:
        list: List of API responses.
    """
    LOGGER.info(
        "Batch request detected. Splitting body and sending individual "
        "requests separately."
    )
    urls = data.split(";")
    service_url = request.base_url
    headers = {"User-Agent": "slo-generator"}
    if "Authorization" in request.headers:
        headers["Authorization"] = request.headers["Authorization"]
        service_url = service_url.replace("http:", "https:")  # force HTTPS auth
    for url in urls:
        if "pubsub_batch_handler" in config:
            LOGGER.info(f"Sending {url} to pubsub batch handler.")
            from google.cloud import pubsub_v1  # pylint: disable=C0415

            # pytype: disable=attribute-error
            # pylint: disable=fixme
            # FIXME `load_config()` returns `Optional[dict]` so `config` can be `None`
            #   so `config` can be `None`
            exporter_conf = config.get("pubsub_batch_handler")
            # pytype: enable=attribute-error
            client = pubsub_v1.PublisherClient()
            project_id = exporter_conf["project_id"]
            topic_name = exporter_conf["topic_name"]
            # pylint: disable=no-member
            topic_path = client.topic_path(project_id, topic_name)
            data = url.encode("utf-8")
            client.publish(topic_path, data=data).result()
        else:  # http
            LOGGER.info(f"Sending {url} to HTTP batch handler.")
            requests.post(service_url, headers=headers, data=url, timeout=10)



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""stubs.py

Stubs for mocking backends and exporters.
"""
import json
import os
import sys
import time
from types import ModuleType

from google.cloud import monitoring_v3

from slo_generator.utils import load_config, load_configs

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
SAMPLE_DIR = os.path.join(os.path.dirname(os.path.dirname(TEST_DIR)), "samples/")

CTX = {
    "PROJECT_ID": "fake",
    "PUBSUB_PROJECT_ID": "fake",
    "PUBSUB_TOPIC_NAME": "fake",
    "GAE_PROJECT_ID": "fake",
    "GAE_MODULE_ID": "fake",
    "GKE_MESH_UID": "fake",
    "GKE_PROJECT_ID": "fake",
    "GKE_CLUSTER_NAME": "fake",
    "GKE_LOCATION": "fake",
    "GKE_SERVICE_NAMESPACE": "fake",
    "GKE_SERVICE_NAME": "fake",
    "LB_PROJECT_ID": "fake",
    "PROMETHEUS_URL": "http://localhost:9090",
    "PROMETHEUS_PUSHGATEWAY_URL": "http://localhost:9091",
    "ELASTICSEARCH_URL": "http://localhost:9200",
    "STACKDRIVER_HOST_PROJECT_ID": "fake",
    "STACKDRIVER_LOG_METRIC_NAME": "fake",
    "BIGQUERY_PROJECT_ID": "fake",
    "BIGQUERY_TABLE_ID": "fake",
    "BIGQUERY_DATASET_ID": "fake",
    "BIGQUERY_TABLE_NAME": "fake",
    "DATADOG_API_KEY": "fake",
    "DATADOG_APP_KEY": "fake",
    "DATADOG_SLO_ID": "fake",
    "DYNATRACE_API_URL": "fake",
    "DYNATRACE_API_TOKEN": "fake",
    "DYNATRACE_SLO_ID": "fake",
}


def add_dynamic(name, code, type):
    """Dynamically add a backend or exporter to slo-generator.

    Args:
        name (str): Name of backend / exporter.
        code (str): Backend / exporter code.
        type (str): 'backends' or 'exporters'.
    """
    mod = ModuleType(name)
    module_name = f"slo_generator.{type}.{name}"
    sys.modules[module_name] = mod
    exec(code, mod.__dict__)


def mock_slo_report(key):
    """Mock SLO report config with edge cases contained in DUMMY_TESTS.

    Args:
        key (str): Key identifying which config to pick from DUMMY_TESTS.

    Returns:
        dict: Dict configuration for SLOReport class.
    """
    slo_config = load_fixture("dummy_slo_config.json")
    ebp_step = load_fixture("dummy_config.json")["error_budget_policies"]["default"][0]
    dummy_tests = load_fixture("dummy_tests.json")
    backend = dummy_tests[key]
    slo_config["spec"]["method"] = backend["method"]
    backend["name"] = "dummy"
    backend["class"] = "Dummy"
    timestamp = time.time()
    return {
        "config": slo_config,
        "backend": backend,
        "step": ebp_step,
        "timestamp": timestamp,
        "client": None,
        "delete": False,
    }


# pylint: disable=too-few-public-methods
class MultiCallableStub:
    """Stub for the grpc.UnaryUnaryMultiCallable interface."""

    def __init__(self, method, channel_stub):
        self.method = method
        self.channel_stub = channel_stub

    # pylint: disable=inconsistent-return-statements
    def __call__(self, request, timeout=None, metadata=None, credentials=None):
        self.channel_stub.requests.append((self.method, request))

        response = None
        if self.channel_stub.responses:
            response = self.channel_stub.responses.pop()

        if isinstance(response, Exception):
            raise response

        if response:
            return response


# pylint: disable=R0903
class ChannelStub:
    """Stub for the grpc.Channel interface."""

    def __init__(self, responses=[]):
        self.responses = responses
        self.requests = []

    # pylint: disable=C0116,W0613
    def unary_unary(self, method, request_serializer=None, response_deserializer=None):
        return MultiCallableStub(method, self)


def mock_grpc_stub(response, proto_method, nresp=1):
    """Fakes gRPC response channel for the proto_method passed.

    Args:
        response (dict): Expected response.
        nresp (int): Number of expected responses.

    Returns:
        ChannelStub: Mocked gRPC channel stub.
    """
    expected_response = proto_method(**response)
    channel = ChannelStub(responses=[expected_response] * nresp)
    return channel


def mock_sd(nresp=1):
    """Fake Stackdriver Monitoring API response for the ListTimeSeries endpoint.

    Args:
        nresp (int): Number of responses to add to response.

    Returns:
        ChannelStub: Mocked gRPC channel stub.
    """
    timeseries = load_fixture("time_series_proto.json")
    response = {"next_page_token": "", "time_series": [timeseries]}
    return mock_grpc_stub(
        response=response,
        proto_method=monitoring_v3.types.ListTimeSeriesResponse,
        nresp=nresp,
    )


# pylint: disable=W0613,R1721
def mock_prom(self, metric):
    """Fake Prometheus query response.

    Args:
        metric (dict): Input metric query.

    Returns:
        dict: Fake response.
    """
    data = {
        "data": {
            "result": [
                {
                    "values": [x for x in range(5)],
                    "value": [0, 1],
                }
            ]
        }
    }
    return json.dumps(data)


# pylint: disable=W0613
def mock_es(self, index, body):
    """Fake ElasticSearch response.

    Args:
        index (str): Index.
        body (dict): Query body.

    Returns:
        dict: Fake response.
    """
    return {"hits": {"total": {"value": 120}}}


def mock_dd_metric_query(*args, **kwargs):
    """Mock Datadog response for datadog.api.Metric.query."""
    return load_fixture("dd_timeseries.json")


def mock_dd_slo_history(*args, **kwargs):
    """Mock Datadog response for datadog.api.ServiceLevelObjective.history."""
    return load_fixture("dd_slo_history.json")


def mock_dd_slo_get(*args, **kwargs):
    """Mock Datadog response for datadog.api.ServiceLevelObjective.get."""
    return load_fixture("dd_slo.json")


def mock_dd_metric_send(*args, **kwargs):
    """Mock Datadog response for datadog.api.Metric.send."""
    return load_fixture("dd_success.json")


def mock_dt(*args, **kwargs):
    """Mock Dynatrace response."""
    if args[0] == "get" and args[1] == "timeseries":
        return load_fixture("dt_metric_get.json")

    elif args[0] == "get" and args[1] == "metrics/query":
        return load_fixture("dt_timeseries_get.json")

    elif args[0] == "get" and args[1].startswith("slo/"):
        return load_fixture("dt_slo_get.json")

    elif args[0] == "post" and args[1] == "entity/infrastructure/custom":
        return load_fixture("dt_metric_send.json")

    elif args[0] == "put" and args[1] == "timeseries":
        return {}


def mock_dt_errors(*args, **kwargs):
    """Mock Dynatrace response with errors."""
    if args[0] == "get" and args[1] == "timeseries":
        return load_fixture("dt_metric_get.json")

    elif args[0] == "get" and args[1] == "metrics/query":
        return load_fixture("dt_timeseries_get.json")

    elif args[0] == "post" and args[1] == "entity/infrastructure/custom":
        return load_fixture("dt_error_rate.json")

    elif args[0] == "put" and args[1] == "timeseries":
        return load_fixture("dt_error_rate.json")


class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def dotize(data):
    """Transform dict to class instance with attribute access.

    Args:
        data (dict): Input dict.

    Returns:
        dotdict: Dotdict equivalent.
    """
    data = dotdict(data)
    for k, v in data.items():
        if isinstance(v, dict):
            data[k] = dotdict(v)
    return data


class mock_ssm_client:
    """Fake Service Monitoring API client."""

    def __init__(self):
        self.services = [dotize(s) for s in load_fixture("ssm_services.json")]
        self.service_level_objectives = [
            dotize(slo) for slo in load_fixture("ssm_slos.json")
        ]

    def common_project_path(self, project_id):
        return f"projects/{project_id}"

    def service_path(self, project_id, service_id):
        project_path = self.project_path(project_id)
        return f"{project_path}/services/{service_id}"

    def create_service(self, parent, service, service_id=None):
        return self.services[0]

    def list_services(self, parent):
        return self.services

    def delete_service(self, name):
        return None

    def create_service_level_objective(
        self, parent, service_level_objective, service_level_objective_id=None
    ):
        return self.service_level_objectives[0]

    def update_service_level_objective(self, service_level_objective):
        return self.service_level_objectives[0]

    def list_service_level_objectives(self, parent):
        return self.service_level_objectives

    def delete_service_level_objective(self, name):
        return None

    @staticmethod
    def to_json(data):
        return data


def get_fixture_path(filename):
    """Get path for a fixture file.

    Args:
        filename (str): Filename of file in fixtures/.

    Returns:
        str: Full path of file in fixtures/.
    """
    return os.path.join(TEST_DIR, "fixtures/", filename)


def load_fixture(filename, ctx=os.environ):
    """Load a fixture from the test/fixtures/ directory and replace context
    environmental variables in it.

    Args:
        filename (str): Filename of the fixture to load.
        ctx (dict): Context dictionary (env variables).

    Returns:
        dict: Loaded fixture.
    """
    path = get_fixture_path(filename)
    return load_config(path, ctx=ctx)


def load_sample(filename, ctx=os.environ):
    """Load a sample from the samples/ directory and replace context
    environmental variables in it.

    Args:
        filename (str): Filename of the fixture to load.
        ctx (dict): Context dictionary (env variables).

    Returns:
        dict: Loaded sample.
    """
    filename = os.path.join(SAMPLE_DIR, filename)
    return load_config(filename, ctx=ctx)


def load_slo_samples(folder_path, ctx=os.environ):
    """List and load all SLO samples from folder path.

    Args:
        folder_path (str): Folder path to load SLO configs from.
        ctx (dict): Context for env variables.

    Returns:
        list: List of loaded SLO configs.
    """
    return load_configs(f"{SAMPLE_DIR}/{folder_path}", ctx)


# Add custom backends / exporters for testing purposes
DUMMY_BACKEND_CODE = open(get_fixture_path("dummy_backend.py")).read()
FAIL_EXPORTER_CODE = open(get_fixture_path("fail_exporter.py")).read()
add_dynamic("dummy", DUMMY_BACKEND_CODE, "backends")
add_dynamic("fail", FAIL_EXPORTER_CODE, "exporters")



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`cloud_monitoring.py`
Cloud Monitoring backend implementation.
"""
import logging
import pprint
import warnings
from collections import OrderedDict

from google.cloud import monitoring_v3

from slo_generator.constants import NO_DATA

LOGGER = logging.getLogger(__name__)


class CloudMonitoringBackend:
    """Backend for querying metrics from Cloud Monitoring.

    Args:
        project_id (str): Cloud Monitoring host project id.
        client (google.cloud.monitoring_v3.MetricServiceClient, optional):
            Existing Cloud Monitoring Metrics client. Initialize a new client
            if omitted.
    """

    def __init__(self, project_id, client=None):
        self.client = client
        if client is None:
            self.client = monitoring_v3.MetricServiceClient()
        self.parent = self.client.common_project_path(project_id)

    # pylint: disable=duplicate-code
    def good_bad_ratio(self, timestamp, window, slo_config):
        """Query two timeseries, one containing 'good' events, one containing
        'bad' events.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window size (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            tuple: A tuple (good_event_count, bad_event_count)
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        filter_good = measurement["filter_good"]
        filter_bad = measurement.get("filter_bad")
        filter_valid = measurement.get("filter_valid")

        # Query 'good events' timeseries
        good_ts = self.query(
            timestamp=timestamp,
            window=window,
            filter=filter_good,
        )
        good_ts = list(good_ts)
        good_event_count = CM.count(good_ts)

        # Query 'bad events' timeseries
        if filter_bad:
            bad_ts = self.query(
                timestamp=timestamp,
                window=window,
                filter=filter_bad,
            )
            bad_ts = list(bad_ts)
            bad_event_count = CM.count(bad_ts)
        elif filter_valid:
            valid_ts = self.query(
                timestamp=timestamp,
                window=window,
                filter=filter_valid,
            )
            valid_ts = list(valid_ts)
            bad_event_count = CM.count(valid_ts) - good_event_count
        else:
            raise Exception("One of `filter_bad` or `filter_valid` is required.")

        LOGGER.debug(
            f"Good events: {good_event_count} | " f"Bad events: {bad_event_count}"
        )

        return good_event_count, bad_event_count

    # pylint: disable=duplicate-code,too-many-locals
    def distribution_cut(self, timestamp, window, slo_config):
        """Query one timeseries of type 'exponential'.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window size (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            tuple: A tuple (good_event_count, bad_event_count).
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        filter_valid = measurement["filter_valid"]
        threshold_bucket = int(measurement["threshold_bucket"])
        good_below_threshold = measurement.get("good_below_threshold", True)

        # Query 'valid' events
        series = self.query(
            timestamp=timestamp,
            window=window,
            filter=filter_valid,
        )
        series = list(series)

        if not series:
            return NO_DATA, NO_DATA  # no timeseries

        distribution_value = series[0].points[0].value.distribution_value
        # bucket_options = distribution_value.bucket_options
        bucket_counts = distribution_value.bucket_counts
        valid_events_count = distribution_value.count
        # growth_factor = bucket_options.exponential_buckets.growth_factor
        # scale = bucket_options.exponential_buckets.scale

        # Explicit the exponential distribution result
        count_sum = 0
        distribution = OrderedDict()
        for i, bucket_count in enumerate(bucket_counts):
            count_sum += bucket_count
            # upper_bound = scale * math.pow(growth_factor, i)
            distribution[i] = {
                # 'upper_bound': upper_bound,
                # 'bucket_count': bucket_count,
                "count_sum": count_sum
            }
        LOGGER.debug(pprint.pformat(distribution))

        if len(distribution) - 1 < threshold_bucket:
            # maximum measured metric is below the cut after bucket number
            lower_events_count = valid_events_count
            upper_events_count = 0
        else:
            lower_events_count = distribution[threshold_bucket]["count_sum"]
            upper_events_count = valid_events_count - lower_events_count

        if good_below_threshold:
            good_event_count = lower_events_count
            bad_event_count = upper_events_count
        else:
            good_event_count = upper_events_count
            bad_event_count = lower_events_count

        return good_event_count, bad_event_count

    def exponential_distribution_cut(self, *args, **kwargs):
        """Alias for `distribution_cut` method to allow for backwards
        compatibility.
        """
        warnings.warn(
            "exponential_distribution_cut will be deprecated in version 2.0, "
            "please use distribution_cut instead",
            FutureWarning,
        )
        return self.distribution_cut(*args, **kwargs)

    # pylint: disable=redefined-builtin,too-many-arguments
    def query(
        self,
        timestamp,
        window,
        filter,
        aligner="ALIGN_SUM",
        reducer="REDUCE_SUM",
        group_by=None,
    ):
        """Query timeseries from Cloud Monitoring.

        Args:
            timestamp (int): Current timestamp.
            window (int): Window size (in seconds).
            filter (str): Query filter.
            aligner (str, optional): Aligner to use.
            reducer (str, optional): Reducer to use.
            group_by (list, optional): List of fields to group by.

        Returns:
            list: List of timeseries objects.
        """
        if group_by is None:
            group_by = []
        measurement_window = CM.get_window(timestamp, window)
        aggregation = CM.get_aggregation(
            window, aligner=aligner, reducer=reducer, group_by=group_by
        )
        request = monitoring_v3.ListTimeSeriesRequest()
        request.name = self.parent
        request.filter = filter
        request.interval = measurement_window
        request.view = monitoring_v3.ListTimeSeriesRequest.TimeSeriesView.FULL
        request.aggregation = aggregation
        timeseries = self.client.list_time_series(request)
        LOGGER.debug(pprint.pformat(timeseries))
        return timeseries

    @staticmethod
    def count(timeseries):
        """Count events in time series assuming it was aligned with ALIGN_SUM
        and reduced with REDUCE_SUM (default).

        Args:
            :obj:`monitoring_v3.TimeSeries`: Timeseries object.

        Returns:
            int: Event count.
        """
        try:
            return timeseries[0].points[0].value.int64_value
        except (IndexError, AttributeError) as exception:
            LOGGER.debug(exception, exc_info=True)
            return NO_DATA  # no events in timeseries

    @staticmethod
    def get_window(timestamp, window):
        """Helper for measurement window.

        Args:
            timestamp (int): Current timestamp.
            window (int): Window size (in seconds).

        Returns:
            :obj:`monitoring_v3.types.TimeInterval`: Measurement window object.
        """
        end_time_seconds = int(timestamp)
        end_time_nanos = int((timestamp - end_time_seconds) * 10**9)
        start_time_seconds = int(timestamp - window)
        start_time_nanos = end_time_nanos
        measurement_window = monitoring_v3.TimeInterval(
            {
                "end_time": {
                    "seconds": end_time_seconds,
                    "nanos": end_time_nanos,
                },
                "start_time": {
                    "seconds": start_time_seconds,
                    "nanos": start_time_nanos,
                },
            }
        )
        LOGGER.debug(pprint.pformat(measurement_window))
        return measurement_window

    @staticmethod
    def get_aggregation(
        window,
        aligner="ALIGN_SUM",
        reducer="REDUCE_SUM",
        group_by=None,
    ):
        """Helper for aggregation object.

        Default aggregation is `ALIGN_SUM`.
        Default reducer is `REDUCE_SUM`.

        Args:
            window (int): Window size (in seconds).
            aligner (str): Aligner type.
            reducer (str): Reducer type.
            group_by (list): List of fields to group by.

        Returns:
            :obj:`monitoring_v3.types.Aggregation`: Aggregation object.
        """
        if group_by is None:
            group_by = []
        aggregation = monitoring_v3.Aggregation(
            {
                "alignment_period": {"seconds": window},
                "per_series_aligner": getattr(
                    monitoring_v3.Aggregation.Aligner, aligner
                ),
                "cross_series_reducer": getattr(
                    monitoring_v3.Aggregation.Reducer, reducer
                ),
                "group_by_fields": group_by,
            }
        )
        LOGGER.debug(pprint.pformat(aggregation))
        return aggregation


CM = CloudMonitoringBackend



# Copyright 2020 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`dynatrace.py`
Datadog backend implementation.
"""
import json
import logging
import pprint

import requests
from retrying import retry

from slo_generator.constants import NO_DATA

LOGGER = logging.getLogger(__name__)


class DynatraceBackend:
    """Backend for querying metrics from Datadog.

    Args:
        client (obj, optional): Existing `requests.Session` to pass.
        api_url (str): Dynatrace API URL.
        api_token (str): Dynatrace token.
    """

    def __init__(self, client=None, api_url=None, api_token=None):
        self.client = client
        if client is None:
            self.client = DynatraceClient(api_url, api_token)

    def query_sli(self, timestamp, window, slo_config):
        """Query SLI value from a given Dynatrace SLO.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            float: SLI value.
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        start = (timestamp - window) * 1000
        end = timestamp * 1000
        slo_id = measurement["slo_id"]
        data = self.retrieve_slo(start, end, slo_id)
        LOGGER.debug(f"Result SLO: {pprint.pformat(data)}")
        sli_value = round(data["evaluatedPercentage"] / 100, 4)
        return sli_value

    def good_bad_ratio(self, timestamp, window, slo_config):
        """Query SLI value from good and valid queries.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            tuple: Good event count, Bad event count.
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        start = (timestamp - window) * 1000
        end = timestamp * 1000
        query_good = measurement["query_good"]
        query_valid = measurement["query_valid"]

        # Good query
        good_event_response = self.query(start=start, end=end, **query_good)
        LOGGER.debug(f"Result good: {pprint.pformat(good_event_response)}")
        good_event_count = DynatraceBackend.count(good_event_response)

        # Good query
        valid_event_response = self.query(start=start, end=end, **query_valid)
        LOGGER.debug(f"Result valid: {pprint.pformat(valid_event_response)}")
        valid_event_count = DynatraceBackend.count(valid_event_response)

        # Return good, bad
        bad_event_count = valid_event_count - good_event_count
        return (good_event_count, bad_event_count)

    def threshold(self, timestamp, window, slo_config):
        """Compute SLI by counting the number of values below and above a
        threshold.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            tuple: Good event count, Bad event count.
        """
        measurement = slo_config["spec"]["service_level_indicator"]
        start = (timestamp - window) * 1000
        end = timestamp * 1000
        query_valid = measurement["query_valid"]
        threshold = measurement["threshold"]
        good_below_threshold = measurement.get("good_below_threshold", True)
        response = self.query(start=start, end=end, **query_valid)
        LOGGER.debug(f"Result valid: {pprint.pformat(response)}")
        return DynatraceBackend.count_threshold(
            response, threshold, good_below_threshold
        )

    # pylint: disable=too-many-arguments
    def query(
        self,
        start,
        end,
        metric_selector=None,
        entity_selector=None,
        aggregation="SUM",
    ):
        """Query Dynatrace Metrics V2.

        Args:
            start (int): Start timestamp (in milliseconds).
            end (int): End timestamp (in milliseconds).
            metric_selector (str): Metric selector.
            entity_selector (str): Entity selector.
            aggregation (str): Aggregation.

        Returns:
            dict: Dynatrace API response.
        """
        params = {
            "from": start,
            "end": end,
            "metricSelector": metric_selector,
            "entitySelector": entity_selector,
            "aggregation": aggregation,
            "includeData": True,
        }
        return self.client.request("get", "metrics/query", version="v2", **params)

    def retrieve_slo(self, start, end, slo_id):
        """Query Dynatrace SLO V2.

        Args:
            start (int): Start timestamp (in milliseconds).
            end (int): End timestamp (in milliseconds).
            slo_id (int): SLO ID.

        Returns:
            dict: Dynatrace API response.
        """
        params = {"from": start, "to": end, "timeFrame": "GTF"}
        endpoint = "slo/" + slo_id
        return self.client.request("get", endpoint, version="v2", **params)

    @staticmethod
    def count(response):
        """Count events in time series data.

        Args:
            response (dict): Dynatrace API response.

        Returns:
            int: Event count.
        """
        try:
            datapoints = response["result"][0]["data"]
            values = []
            for point in datapoints:
                point_values = [
                    point
                    for point in point["values"]
                    if point is not None and point > 0
                ]
                values.extend(point_values)
            return sum(values)
        except (IndexError, KeyError) as exception:
            LOGGER.warning("Couldn't find any values in timeseries response")
            LOGGER.debug(exception)
            return NO_DATA  # no events in timeseries

    @staticmethod
    def count_threshold(response, threshold, good_below_threshold=True):
        """Create 2 buckets based on response and a value threshold, and return
        number of events contained in each bucket.

        Args:
            response (dict): Dynatrace API response.
            threshold (int): Threshold.
            good_below_threshold (bool): If true, good events are < threshold.

        Returns:
            tuple: Number of good events, Number of bad events.
        """
        try:
            datapoints = response["result"][0]["data"]
            below = []
            above = []
            for point in datapoints:
                points_below = [
                    point
                    for point in point["values"]
                    if point is not None and point < threshold
                ]
                points_above = [
                    point
                    for point in point["values"]
                    if point is not None and point > threshold
                ]
                below.extend(points_below)
                above.extend(points_above)
            if good_below_threshold:
                return len(below), len(above)
            return len(above), len(below)
        except (IndexError, KeyError, ZeroDivisionError) as exception:
            LOGGER.warning("Couldn't find any values in timeseries response")
            LOGGER.debug(exception)
            return NO_DATA, NO_DATA  # no events in timeseries


def retry_http(response):
    """Retry on specific HTTP errors:

        * 429: Rate limited to 50 reqs/minute.

    Args:
        response (dict): Dynatrace API response.

    Returns:
        bool: True to retry, False otherwise.
    """
    retry_codes = [429]
    returned_code = response.get("error", {})
    if isinstance(returned_code, str):
        code = 200
    else:
        code = int(returned_code.get("code", 200))
    return code in retry_codes


class DynatraceClient:
    """Small wrapper around requests to query Dynatrace API.

    Args:
        api_url (str): Dynatrace API URL.
        api_token (str): Dynatrace token.
    """

    # Keys to extract response data for each endpoint
    ENDPOINT_KEYS = {"metrics": "metrics", "metrics/query": "result"}

    def __init__(self, api_url, api_key):
        self.client = requests.Session()
        self.url = api_url.rstrip("/")
        self.token = api_key

    @retry(
        retry_on_result=retry_http,
        wait_exponential_multiplier=1000,
        wait_exponential_max=10000,
        stop_max_delay=10000,
    )
    # pylint: disable=too-many-arguments,too-many-locals
    def request(
        self,
        method,
        endpoint,
        name=None,
        version="v1",
        post_data=None,
        key=None,
        **params,
    ):
        """Request Dynatrace API.

        Args:
            method (str): Requests method between ['post', 'put', 'get'].
            endpoint (str): API endpoint.
            name (str): API resource name.
            version (str): API version. Default: v1.
            post_data (dict): JSON data.
            key (str): Key to extract data from JSON response.
            params (dict): Params to send with request.

        Returns:
            obj: API response.
        """
        req = getattr(self.client, method)
        url = f"{self.url}/api/{version}/{endpoint}"
        params["Api-Token"] = self.token
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": "slo-generator",
        }
        if name:
            url += f"/{name}"
        params_str = "&".join(
            f"{key}={val}" for key, val in params.items() if val is not None
        )
        url += f"?{params_str}"
        LOGGER.debug(f'Running "{method}" request to {url} ...')
        if method in ["put", "post"]:
            response = req(url, headers=headers, json=post_data)
        else:
            response = req(url, headers=headers)
        LOGGER.debug(f"Response: {response}")
        data = DynatraceClient.to_json(response)
        next_page_key = data.get("nextPageKey")
        if next_page_key:
            params = {"nextPageKey": next_page_key, "Api-Token": self.token}
            LOGGER.debug(f"Requesting next page: {next_page_key}")
            data_next = self.request(method, endpoint, name, version, **params)
            next_page_key = data_next.get("nextPageKey")
            if not key:
                key = DynatraceClient.ENDPOINT_KEYS.get(endpoint, "result")
            data[key].extend(data_next[key])
        return data

    @staticmethod
    def to_json(resp):
        """Decode JSON response from Python requests response as utf-8 and
        replace \n characters.

        Args:
            resp (requests.Response): API response.

        Returns:
            dict: API JSON response.
        """
        res = resp.content.decode("utf-8").replace("\n", "")
        data = json.loads(res)
        return data



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`cloud_service_monitoring.py`
Cloud Service Monitoring exporter class.
"""
import difflib
import json
import logging
import os
import warnings
from typing import Optional, Sequence, Union

import google.api_core.exceptions
from google.cloud.monitoring_v3 import ServiceMonitoringServiceClient

# pytype: disable=pyi-error
from google.protobuf.json_format import MessageToJson

from slo_generator.backends.cloud_monitoring import CloudMonitoringBackend
from slo_generator.constants import NO_DATA
from slo_generator.utils import dict_snake_to_caml

# pytype: enable=pyi-error

LOGGER = logging.getLogger(__name__)

SID_GAE: str = "gae:{project_id}_{module_id}"
SID_CLOUD_ENDPOINT: str = "ist:{project_id}-{service}"
SID_CLUSTER_ISTIO: str = (
    "ist:{project_id}-{suffix}-{location}-{cluster_name}-{service_namespace}-"
    "{service_name}"
)
SID_MESH_ISTIO: str = "ist:{mesh_uid}-{service_namespace}-{service_name}"


# pylint: disable=too-many-public-methods
class CloudServiceMonitoringBackend:
    """Cloud Service Monitoring backend class.

    Args:
        project_id (str): Cloud Monitoring host project id.
        client (google.cloud.monitoring_v3.ServiceMonitoringServiceClient):
            Existing Service Monitoring API client. Initialize a new client if
            omitted.
    """

    def __init__(self, project_id: str, client=None):
        self.project_id = project_id
        self.client = client
        if client is None:
            self.client = ServiceMonitoringServiceClient()
        self.parent = self.client.common_project_path(project_id)
        self.workspace_path = f"workspaces/{project_id}"
        self.project_path = f"projects/{project_id}"

    def good_bad_ratio(self, timestamp: int, window: int, slo_config: dict) -> tuple:
        """Good bad ratio method.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            tuple: SLO config.
        """
        return self.retrieve_slo(timestamp, window, slo_config)

    def distribution_cut(self, timestamp: int, window: int, slo_config: dict) -> tuple:
        """Distribution cut method.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            tuple: SLO config.
        """
        return self.retrieve_slo(timestamp, window, slo_config)

    def basic(self, timestamp: int, window: int, slo_config: dict) -> tuple:
        """Basic method (automatic SLOs for GAE / GKE (Istio) and Cloud
        Endpoints).

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            tuple: SLO config.
        """
        return self.retrieve_slo(timestamp, window, slo_config)

    def window(self, timestamp: int, window: int, slo_config: dict) -> tuple:
        """Window-based SLI method.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            tuple: SLO config.
        """
        return self.retrieve_slo(timestamp, window, slo_config)

    # pylint: disable=unused-argument
    def delete(self, timestamp: int, window: int, slo_config: dict) -> Optional[dict]:
        """Delete method.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            dict: SLO config.
        """
        return self.delete_slo(window, slo_config)

    def retrieve_slo(self, timestamp: int, window: int, slo_config: dict):
        """Get SLI value from Cloud Monitoring API.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            dict: SLO config.
        """
        # Get or create service
        service = self.get_service(slo_config)
        if service is None:
            service = self.create_service(slo_config)
        LOGGER.debug(service)

        # Get or create SLO
        slo = self.get_slo(window, slo_config)
        if not slo:
            slo = self.create_slo(window, slo_config)
        LOGGER.debug(service)

        # Now that we have our SLO, retrieve the TimeSeries from Cloud
        # Monitoring API for that particular SLO id.
        metric_filter = self.build_slo_id(window, slo_config, full=True)
        # pylint: disable=redefined-builtin
        filter = f'select_slo_counts("{metric_filter}")'

        # Query SLO timeseries
        cloud_monitoring = CloudMonitoringBackend(self.project_id)
        timeseries = cloud_monitoring.query(
            timestamp,
            window,
            filter,
            aligner="ALIGN_SUM",
            reducer="REDUCE_SUM",
            group_by=["metric.labels.event_type"],
        )
        timeseries = list(timeseries)
        good_event_count, bad_event_count = SSM.count(timeseries)
        return (good_event_count, bad_event_count)

    @staticmethod
    def count(timeseries: list):
        """Extract good_count, bad_count tuple from Cloud Monitoring API
        response.

        Args:
            timeseries (list): List of timeseries objects.

        Returns:
            tuple: A tuple (good_event_count, bad_event_count).
        """
        good_event_count, bad_event_count = NO_DATA, NO_DATA
        for timeserie in timeseries:
            event_type = timeserie.metric.labels["event_type"]
            value = timeserie.points[0].value.double_value
            if event_type == "bad":
                bad_event_count = value
            elif event_type == "good":
                good_event_count = value
        return good_event_count, bad_event_count

    def create_service(self, slo_config: dict) -> dict:
        """Create Service object in Cloud Service Monitoring API.

        Args:
            slo_config (dict): SLO configuration.

        Returns:
            dict: Cloud Service Monitoring API response.
        """
        LOGGER.debug("Creating service ...")
        service_json = self.build_service(slo_config)
        service_id = self.build_service_id(slo_config)
        service = self.client.create_service(
            request={
                "parent": self.project_path,
                "service": service_json,
                "service_id": service_id,
            }
        )
        LOGGER.info(
            f'Service "{service_id}" created successfully in Cloud '
            f"Service Monitoring API."
        )
        return SSM.to_json(service)

    def get_service(self, slo_config: dict) -> Optional[dict]:
        """Get Service object from Cloud Service Monitoring API.

        Args:
            slo_config (dict): SLO configuration.

        Returns:
            dict: Service config.
        """

        # Look for API services in workspace matching our config.
        service_id = self.build_service_id(slo_config)
        services = list(
            self.client.list_services(
                request={
                    "parent": self.workspace_path,
                }
            )
        )
        matches = [
            service for service in services if service.name.split("/")[-1] == service_id
        ]

        # If no match is found for our service name in the API, raise an
        # exception if the service should have been auto-added (method 'basic'),
        # else output a warning message.
        if not matches:
            msg = (
                f'Service "{service_id}" does not exist in '
                f'workspace "{self.project_id}"'
            )
            method = slo_config["spec"]["method"]
            if method == "basic":
                sids = [service.name.split("/")[-1] for service in services]
                LOGGER.debug(f"List of services in workspace {self.project_id}: {sids}")
                raise Exception(msg)
            LOGGER.error(msg)
            return None

        # Match found in API, return it.
        service = matches[0]
        LOGGER.debug(f'Found matching service "{service.name}"')
        return SSM.to_json(service)

    def build_service(self, slo_config: dict) -> dict:
        """Build service JSON in Cloud Monitoring API from SLO
        configuration.

        Args:
            slo_config (dict): SLO configuration.

        Returns:
            dict: Service JSON in Cloud Monitoring API.
        """
        service_id = self.build_service_id(slo_config)
        display_name = slo_config.get("service_display_name", service_id)
        return {"display_name": display_name, "custom": {}}

    def build_service_id(
        self,
        slo_config: dict,
        dest_project_id: Optional[str] = None,
        full: bool = False,
    ):
        """Build service id from SLO configuration.

        Args:
            slo_config (dict): SLO configuration.
            dest_project_id (str, optional): Project id for service if different
                than the workspace project id.
            full (bool): If True, return full service resource id including
                project path.

        Returns:
            str: Service id.
        """
        project_id = self.project_id
        measurement = slo_config["spec"]["service_level_indicator"]
        app_engine = measurement.get("app_engine")
        cluster_istio = measurement.get("cluster_istio")
        mesh_istio = measurement.get("mesh_istio")
        cloud_endpoints = measurement.get("cloud_endpoints")

        # Use auto-generated ids for 'custom' SLOs, use system-generated ids
        # for all other types of SLOs.
        if app_engine:
            service_id = SID_GAE.format_map(app_engine)
            dest_project_id = app_engine["project_id"]
        elif cluster_istio:
            warnings.warn(
                "ClusterIstio is deprecated in the Service Monitoring API."
                "It will be removed in version 3.0, please use MeshIstio "
                "instead",
                FutureWarning,
            )
            if "zone" in cluster_istio:
                cluster_istio["suffix"] = "zone"
                cluster_istio["location"] = cluster_istio["zone"]
            elif "location" in cluster_istio:
                cluster_istio["suffix"] = "location"
            service_id = SID_CLUSTER_ISTIO.format_map(cluster_istio)
            dest_project_id = cluster_istio["project_id"]
        elif mesh_istio:
            service_id = SID_MESH_ISTIO.format_map(mesh_istio)
        elif cloud_endpoints:
            service_id = SID_CLOUD_ENDPOINT.format_map(cloud_endpoints)
            dest_project_id = cluster_istio["project_id"]
        else:  # user-defined service id
            service_name = slo_config["metadata"]["labels"].get("service_name", "")
            feature_name = slo_config["metadata"]["labels"].get("feature_name", "")
            service_id = slo_config["spec"]["service_level_indicator"].get("service_id")
            if not service_id:
                if not service_name or not feature_name:
                    raise Exception(
                        "Service id not set in SLO configuration. Please set "
                        "either `spec.service_level_indicator.service_id` or "
                        "both `metadata.labels.service_name` and "
                        "`metadata.labels.feature_name` in your SLO "
                        "configuration."
                    )
                service_id = f"{service_name}-{feature_name}"

        if full:
            if dest_project_id:
                return f"projects/{dest_project_id}/services/{service_id}"
            return f"projects/{project_id}/services/{service_id}"

        return service_id

    def create_slo(self, window: int, slo_config: dict) -> dict:
        """Create SLO object in Cloud Service Monitoring API.

        Args:
            window (int): Window (in seconds).
            slo_config (dict): SLO config.

        Returns:
            dict: Service Management API response.
        """
        slo_json = SSM.build_slo(window, slo_config)
        slo_id = self.build_slo_id(window, slo_config)
        parent = self.build_service_id(slo_config, full=True)
        slo = self.client.create_service_level_objective(
            request={
                "parent": parent,
                "service_level_objective": slo_json,
                "service_level_objective_id": slo_id,
            }
        )
        return SSM.to_json(slo)

    # pylint: disable=R0912,R0915
    @staticmethod
    # pylint: disable=R0912,R0915,too-many-locals
    def build_slo(window: int, slo_config: dict) -> dict:
        """Get SLO JSON representation in Cloud Service Monitoring API from SLO
        configuration.

        Args:
            window (int): Window (in seconds).
            slo_config (dict): SLO Configuration.

        Returns:
            dict: SLO JSON configuration.
        """
        measurement = slo_config["spec"].get("service_level_indicator", {})
        method = slo_config["spec"]["method"]
        description = slo_config["spec"]["description"]
        goal = slo_config["spec"]["goal"]
        minutes, _ = divmod(window, 60)
        hours, _ = divmod(minutes, 60)
        display_name = f"{description} ({hours}h)"
        slo = {
            "display_name": display_name,
            "goal": goal,
            "rolling_period": {"seconds": window},
        }
        filter_valid = measurement.get("filter_valid", "")
        if method == "basic":
            methods = measurement.get("method", [])
            locations = measurement.get("location", [])
            versions = measurement.get("version", [])
            threshold = measurement.get("latency", {}).get("threshold")
            slo["service_level_indicator"] = {"basic_sli": {}}
            basic_sli = slo["service_level_indicator"]["basic_sli"]
            if methods:
                basic_sli["method"] = methods
            if locations:
                basic_sli["location"] = locations
            if versions:
                basic_sli["version"] = versions
            if threshold:
                basic_sli["latency"] = {
                    "threshold": {
                        "seconds": 0,
                        "nanos": int(threshold) * 10**6,
                    }
                }
            else:
                basic_sli["availability"] = {}

        elif method == "good_bad_ratio":
            filter_good = measurement.get("filter_good", "")
            filter_bad = measurement.get("filter_bad", "")
            slo["service_level_indicator"] = {
                "request_based": {
                    "good_total_ratio": {},
                }
            }
            sli = slo["service_level_indicator"]
            ratio = sli["request_based"]["good_total_ratio"]
            if filter_good:
                ratio["good_service_filter"] = filter_good
            if filter_bad:
                ratio["bad_service_filter"] = filter_bad
            if filter_valid:
                ratio["total_service_filter"] = filter_valid

        elif method == "distribution_cut":
            range_min = measurement.get("range_min", 0)
            range_max = measurement["range_max"]
            slo["service_level_indicator"] = {
                "request_based": {
                    "distribution_cut": {
                        "distribution_filter": filter_valid,
                        "range": {
                            "max": float(range_max),
                        },
                    }
                }
            }
            sli = slo["service_level_indicator"]["request_based"]
            if range_min != 0:
                sli["distribution_cut"]["range"]["min"] = float(range_min)

        elif method == "windows":
            # pylint: disable=redefined-builtin
            filter = measurement.get("filter")
            # threshold = conf.get('threshold')
            # mean_in_range = conf.get('filter')
            # sum_in_range = conf.get('filter')
            slo["service_level_indicator"] = {
                "windows_based": {
                    "window_period": window,
                    "good_bad_metric_filter": filter,
                    # 'good_total_ratio_threshold': {
                    #   object (PerformanceThreshold)
                    # },
                    # 'metricMeanInRange': {
                    #   object (MetricRange)
                    # },
                    # 'metricSumInRange': {
                    #   object (MetricRange)
                    # }
                }
            }
        else:
            raise Exception(f'Method "{method}" is not supported.')
        return slo

    def get_slo(self, window: int, slo_config: dict) -> Optional[dict]:
        """Get SLO object from Cloud Service Monssitoring API.

        Args:
            window (int): Window in seconds.
            slo_config (dict): SLO config.

        Returns:
            dict: API response.
        """
        service_path = self.build_service_id(slo_config, full=True)
        LOGGER.debug(f'Getting SLO for for "{service_path}" ...')
        slos = self.list_slos(service_path)
        slo_local_id = self.build_slo_id(window, slo_config)
        slo_json = SSM.build_slo(window, slo_config)
        slo_json = SSM.convert_slo_to_ssm_format(slo_json)

        # Loop through API response to find an existing SLO that corresponds to
        # our configuration.
        for slo in slos:
            slo_remote_id = slo["name"].split("/")[-1]
            equal = slo_remote_id == slo_local_id
            if equal:
                LOGGER.debug(f'Found existing SLO "{slo_remote_id}".')
                LOGGER.debug(f"SLO object: {slo}")
                strict_equal = SSM.compare_slo(slo_json, slo)
                if strict_equal:
                    return slo
                return self.update_slo(window, slo_config)
        LOGGER.warning("No SLO found matching configuration.")
        LOGGER.debug(f"SLOs from Cloud Service Monitoring API: {slos}")
        LOGGER.debug(f"SLO config converted: {slo_json}")
        return None

    def update_slo(self, window: int, slo_config: dict) -> dict:
        """Update an existing SLO.

        Args:
            window (int): Window (in seconds)
            slo_config (dict): SLO configuration.

        Returns:
            dict: API response.
        """
        slo_json = SSM.build_slo(window, slo_config)
        slo_id = self.build_slo_id(window, slo_config, full=True)
        LOGGER.warning(f"Updating SLO {slo_id} ...")
        slo_json["name"] = slo_id
        return SSM.to_json(
            self.client.update_service_level_objective(
                request={
                    "service_level_objective": slo_json,
                }
            )
        )

    def list_slos(self, service_path: str) -> list:
        """List all SLOs from Cloud Service Monitoring API.

        Args:
            service_path (str): Service path in the form
                'projects/{project_id}/services/{service_id}'.

        Returns:
            list: API response.
        """
        slos = self.client.list_service_level_objectives(
            request={
                "parent": service_path,
            }
        )
        slos = list(slos)
        LOGGER.debug(f"{len(slos)} SLOs found in Cloud Service Monitoring API.")
        # LOGGER.debug(slos)
        return [SSM.to_json(slo) for slo in slos]

    def delete_slo(self, window: int, slo_config: dict) -> Optional[dict]:
        """Delete SLO from Cloud Service Monitoring API.

        Args:
            window (int): Window (in seconds).
            slo_config: SLO configuration.

        Returns:
            dict: API response.
        """
        slo_path = self.build_slo_id(window, slo_config, full=True)
        LOGGER.info(f'Deleting SLO "{slo_path}"')
        try:
            return self.client.delete_service_level_objective(
                request={
                    "name": slo_path,
                }
            )
        except google.api_core.exceptions.NotFound:
            LOGGER.warning(
                f'SLO "{slo_path}" does not exist in Service Monitoring API. '
                f"Skipping."
            )
            return None

    def build_slo_id(self, window: int, slo_config: dict, full: bool = False) -> str:
        """Build SLO id from SLO configuration.

        Args:
            slo_config (dict): SLO configuration.
            full (bool): If True, return full resource id including project.

        Returns:
            str: SLO id.
        """
        sli = slo_config["spec"]["service_level_indicator"]
        slo_name = slo_config["metadata"]["labels"].get("slo_name")
        slo_id = sli.get("slo_id", slo_name)
        if not slo_id:
            raise Exception(
                "SLO id not set in SLO configuration. Please set either "
                "`spec.service_level_indicator.slo_id` or "
                "`metadata.labels.slo_name` in your SLO configuration."
            )
        full_slo_id = f"{slo_id}-{window}"
        if full:
            service_path = self.build_service_id(slo_config, full=True)
            return f"{service_path}/serviceLevelObjectives/{full_slo_id}"
        return full_slo_id

    @staticmethod
    def compare_slo(slo1: dict, slo2: dict) -> bool:
        """Compares 2 SLO configurations to see if they correspond to the same
        SLO.

        An SLO is deemed the same if the whole configuration is similar, except
        for the `goal` field that should be adjustable.

        Args:
            slo1 (dict): Service Monitoring API SLO configuration to compare.
            slo2 (dict): Service Monitoring API SLO configuration to compare.

        Returns:
            bool: True if the SLOs match, False otherwise.
        """
        exclude_keys = ["name"]
        slo1_copy = {k: v for k, v in slo1.items() if k not in exclude_keys}
        slo2_copy = {k: v for k, v in slo2.items() if k not in exclude_keys}
        local_json = json.dumps(slo1_copy, sort_keys=True)
        remote_json = json.dumps(slo2_copy, sort_keys=True)
        if os.environ.get("DEBUG") == "2":
            LOGGER.info("----------")
            LOGGER.info(local_json)
            LOGGER.info("----------")
            LOGGER.info(remote_json)
            LOGGER.info("----------")
            LOGGER.info(SSM.string_diff(local_json, remote_json))
        return local_json == remote_json

    @staticmethod
    def string_diff(
        string1: Union[str, Sequence[str]], string2: Union[str, Sequence[str]]
    ) -> list:
        """Diff 2 strings. Used to print comparison of JSONs for debugging.

        Args:
            string1 (str): String 1.
            string2 (str): String 2.

        Returns:
            list: List of messages pointing out differences.
        """
        lines = []
        for idx, string in enumerate(difflib.ndiff(string1, string2)):
            if string[0] == " ":
                continue
            last = string[-1]
            if string[0] == "-":
                info = f'Delete "{last}" from position {idx}'
                lines.append(info)
            elif string[0] == "+":
                info = f'Add "{last}" to position {idx}'
                lines.append(info)
        return lines

    @staticmethod
    def convert_slo_to_ssm_format(slo: dict) -> dict:
        """Convert SLO JSON to Cloud Service Monitoring API format.
        Address edge cases, like `duration` object computation.

        Args:
            slo (dict): SLO JSON object to be converted to Cloud Service
                Monitoring API format.

        Returns:
            dict: SLO configuration in Cloud Service Monitoring API format.
        """
        # Our local JSON is in snake case, convert it to Caml case.
        data = dict_snake_to_caml(slo)

        # The `rollingPeriod` field is in Duration format, convert it.
        try:
            period = data["rollingPeriod"]
            data["rollingPeriod"] = SSM.convert_duration_to_string(period)
        except KeyError:
            pass

        # The `latency` field is in Duration format, convert it.
        try:
            latency = data["serviceLevelIndicator"]["basicSli"]["latency"]
            threshold = latency["threshold"]
            latency["threshold"] = SSM.convert_duration_to_string(threshold)
        except KeyError:
            pass

        return data

    @staticmethod
    def convert_duration_to_string(duration):
        """Convert a duration object to a duration string (in seconds).

        Args:
            duration (dict): Duration dictionary.

        Returns:
            str: Duration string.
        """
        duration_seconds = 0.000
        if "seconds" in duration:
            duration_seconds += duration["seconds"]
        if "nanos" in duration:
            duration_seconds += duration["nanos"] * 10 ** (-9)
        if duration_seconds.is_integer():
            duration_str = int(duration_seconds)
        else:
            duration_str = f"{duration_seconds:0.3f}"
        return str(duration_str) + "s"

    @staticmethod
    def to_json(response):
        """Convert a Cloud Service Monitoring API response to JSON
        format.

        Args:
            response (obj): Response object.

        Returns:
            dict: Response object serialized as JSON.
        """
        # pylint: disable=protected-access
        return json.loads(MessageToJson(response._pb))


SSM = CloudServiceMonitoringBackend



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from slo_generator.migrations.migrator import slo_config_v1tov2

from .test_stubs import load_fixture


class TestMigrator(unittest.TestCase):
    def setUp(self):
        self.slo_config_v1 = load_fixture("slo_config_v1.yaml")
        self.slo_config_v2 = load_fixture("slo_config_v2.yaml")
        self.shared_config = {
            "backends": {},
            "exporters": {},
            "error_budget_policies": {},
        }

    def test_migrate_v1_to_v2(self):
        slo_config_migrated = slo_config_v1tov2(
            self.slo_config_v1, self.shared_config, quiet=True
        )
        self.assertDictEqual(slo_config_migrated, self.slo_config_v2)
